#define MAX(x,y) ((x)>(y)?(x):(y))
#define MIN(x,y) ((x)<(y)?(x):(y))

extern "C" void
cuda_assign_inv_mult_eoprec_wd_bicgstab_(double   *kappa,
                                         double   *tol,
                                         int      *iter,
                                         f_wqf_eo *yde,
                                         f_wqf_eo *ye) {
//
// MAIN GPU Solver with BiCGStab algorithm for 
//  v
//  Dee yde = ye
//        v
// with   Dee = 1 - kappa^2 Fee Meo Foo Moe
//

  //===========================================================
  // host variables
  //===========================================================
#define NOERROR         (0)
#define ERROR_BREAKDOWN (1)
  int ierr_flag;
  const int niter = 10000;
  int ieo;
  const float tol_bk = 1.0e-14;
  const float delta = 0.01f;
  float mkappa2 = -(float)((*kappa)*(*kappa));
  float    tol2 =  (float)((*tol)*(*tol));
  float maxerr,minerr;
  float err;
  float rtmp0;
  cuComplex  alp, bet,  omg, rho0, rho1, ctmp0, ctmp1;
  cuComplex nalp,      nomg;

  //===========================================================
  // device variables
  //===========================================================
  float_gpu      mkappa2d;
  cuComplex_gpu  alpd, betd, omgd;
  cuComplex_gpu nalpd,      nomgd;
  cu_wqf_eo *yed,*yded,*red,*rted,*ped,*qed,*ted,*ved;
  cu_wqf_eo *xmind;

  ierr_flag = NOERROR;

#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    switch_gpu_(&igx);
#if CUDA_VERSION >= 4000
    CUDA_SAFE_CALL(cudaDeviceSynchronize());
#else
    CUDA_SAFE_CALL(cudaThreadSynchronize());
#endif
  }

  startTimer(solv_timer);

#ifdef _DEBUG_
  if ( 0 == _nodeid_ ) {
    fflush(stdout);
    if (is_clover) {
      printf("%%S_BICGSTAB_CLOVER(MPI+GPU-PARALLEL)\n");
    } else {
      printf("%%S_BICGSTAB_WILSON(MPI+GPU-PARALLEL)\n");
    }
    fflush(stdout);
  }
#endif

   ieo = (*ye).ieo;

   yed = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
  yded = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   red = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
  rted = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   ped = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   qed = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   ted = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   ved = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   xmind = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);

   (*yed).ieo = ieo;
  (*yded).ieo = ieo;
   (*red).ieo = ieo;
  (*rted).ieo = ieo;
   (*ped).ieo = ieo;
   (*qed).ieo = ieo;
   (*ted).ieo = ieo;
   (*ved).ieo = ieo;
   (*xmind).ieo = ieo;

  (*yded).ipeo = _ipeo_;
   (*yed).ipeo = _ipeo_;
   (*red).ipeo = _ipeo_;
  (*rted).ipeo = _ipeo_;
   (*ped).ipeo = _ipeo_;
   (*qed).ipeo = _ipeo_;
   (*ted).ipeo = _ipeo_;
   (*ved).ipeo = _ipeo_;
   (*xmind).ipeo = _ipeo_;

  //======================================
  // allocate device memory
  //======================================
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    switch_gpu_(&igx);

    CUDA_SAFE_CALL(cudaMalloc((void**)    &(alpd.g[igx]),sizeof(cuComplex)));
    CUDA_SAFE_CALL(cudaMalloc((void**)   &(nalpd.g[igx]),sizeof(cuComplex)));
    CUDA_SAFE_CALL(cudaMalloc((void**)    &(betd.g[igx]),sizeof(cuComplex)));
    CUDA_SAFE_CALL(cudaMalloc((void**)    &(omgd.g[igx]),sizeof(cuComplex)));
    CUDA_SAFE_CALL(cudaMalloc((void**)   &(nomgd.g[igx]),sizeof(cuComplex)));

    CUDA_SAFE_CALL(cudaMalloc((void**)&(mkappa2d.g[igx]),sizeof(float)));

    CUDA_SAFE_CALL(cudaMalloc((void**) &((*yed).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)&((*yded).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**) &((*red).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)&((*rted).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**) &((*ped).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**) &((*qed).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**) &((*ted).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**) &((*ved).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**) &((*xmind).g[igx]) ,sizeof(cu_wqf_eo_blk)));

    new_q((*yded).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*yed).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*red).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q((*rted).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*ped).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*qed).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*ted).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*ved).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
  }


#ifdef _DEBUG_
  if ( 0 == _nodeid_ ) {
    printf("%% cudaMalloc OK.\n");
  }
#endif

  //======================================
  // Send data from host to device
  //======================================
  startTimer(ysend_timer);
  assign(mkappa2d,mkappa2);
  assign(yed,ye);
  stopTimer(ysend_timer);

  //
  // set x_0 = 0 (yde = 0)
  //
  clear(yded,ieo);

  //
  // r  = b ( = y : rhs of equation )
  // rt = r
  // p  = r
  //
  assign( red,yed);
  assign(rted,red);
  assign( ped,red);

  //
  // rho0 = |re|^2
  //
   rtmp0 = abs2(red);
  rho0.x = rtmp0;
  rho0.y = 0.0f;
  solv_flop_count += QABS2;

  err = sqrt(rtmp0);
  maxerr = err;
  minerr = err;
#ifdef _DEBUG_
  err = sqrt(rtmp0);
  if ( 0 == _nodeid_ ) printf("%% NRAS=%3d NMR=%3d NDEPTH=%3d NGX=%3d\n",_NRAS,_NMR,_NDEPTH,_NGX);
  if ( 0 == _nodeid_ ) printf("%% S_TOL=%24.15e\n",sqrt(tol2));
  if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
#endif

  while (*iter < niter) {

    //
    // v = M p 
    // q = A v 
    //
    if (is_clover) {

#ifdef _PREC_
      assign_inv_mult_eoprec_wd_ras(mkappa2d,ved,ped,ud,fclinvd);
      solv_flop_count += MULT_PREC;
           assign_mult_eoprec_wd_eo(mkappa2d,qed,ved,ud,fclinvd);
#else
           assign_mult_eoprec_wd_eo(mkappa2d,qed,ped,ud,fclinvd);
#endif

    } else {

#ifdef _PREC_
      assign_inv_mult_eoprec_wd_ras(mkappa2d,ved,ped,ud);
      solv_flop_count += MULT_PREC;
           assign_mult_eoprec_wd_eo(mkappa2d,qed,ved,ud);
#else
           assign_mult_eoprec_wd_eo(mkappa2d,qed,ped,ud);
#endif

    }

    solv_flop_count += MULT_MB_PRE;
    *iter = *iter + 1;

    //
    // alp = rho0/<rt|q>
    //
    ctmp0 = prod(rted,qed);
    if ( sqrt(ctmp0.x*ctmp0.x + ctmp0.y*ctmp0.y) < tol_bk) {
      ierr_flag = ERROR_BREAKDOWN;
      break;
    }

      alp = assign_div(rho0,ctmp0);
    if ( sqrt(alp.x*alp.x + alp.y*alp.y) < tol_bk) {
      ierr_flag = ERROR_BREAKDOWN;
      break;
    }
    assign( alpd, alp);
     nalp = assign_neg(alp);
    assign(nalpd,nalp);

    solv_flop_count += QPROD;

#ifdef _DEBUG_
    if ( 0 == _nodeid_ ) {
      printf("  <rt|q> = %24.15e%24.15e\n",ctmp0.x,ctmp0.y);
      printf("    rho0 = %24.15e%24.15e\n",rho0.x,rho0.y);
      printf("     alp = %24.15e%24.15e\n",alp.x,alp.y);
    }
#endif

    //
    // x = x + alp v
    // r = r - alp q
    //
#ifdef _PREC_
    accum_add_mult(yded, alpd,ved);
#else
    accum_add_mult(yded, alpd,ped);
#endif
    accum_add_mult( red,nalpd,qed);

    solv_flop_count += QACCUM_ADD_CMULT*2;

    rtmp0 = abs2(red);
#ifdef _DEBUG_
    err = sqrt(rtmp0);
    if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    solv_flop_count += QABS2;

    err = sqrt(rtmp0);
    maxerr = MAX(maxerr,err);
    if ( err < minerr) {
      assign(xmind,yded);
      minerr = err;
    }

    if (rtmp0 < tol2) break;

    //
    // v = M r
    // t = A v
    //
    if (is_clover) {
#ifdef _PREC_
      assign_inv_mult_eoprec_wd_ras(mkappa2d,ved,red,ud,fclinvd);
      solv_flop_count += MULT_PREC;
           assign_mult_eoprec_wd_eo(mkappa2d,ted,ved,ud,fclinvd);
#else
           assign_mult_eoprec_wd_eo(mkappa2d,ted,red,ud,fclinvd);
#endif
    } else {
#ifdef _PREC_
      assign_inv_mult_eoprec_wd_ras(mkappa2d,ved,red,ud);
      solv_flop_count += MULT_PREC;
           assign_mult_eoprec_wd_eo(mkappa2d,ted,ved,ud);
#else
           assign_mult_eoprec_wd_eo(mkappa2d,ted,red,ud);
#endif
    }
    solv_flop_count += MULT_MB_PRE;
    *iter = *iter + 1;

    //
    // omg = <t|r>/<t|t>
    //
    rtmp0 = abs2(ted);
    if ( rtmp0 < tol_bk ) {
      ierr_flag = ERROR_BREAKDOWN;
      break;
    }

    ctmp0 = prod(ted,red);
      omg = assign_div(ctmp0,rtmp0);
    if ( sqrt(omg.x*omg.x + omg.y*omg.y) < tol_bk) {
      ierr_flag = ERROR_BREAKDOWN;
      break;
    }

    assign( omgd, omg);
     nomg = assign_neg(omg);
    assign(nomgd,nomg);
    solv_flop_count += QABS2 + QPROD;

#ifdef _DEBUG_
    if ( 0 == _nodeid_ ) {
      printf("   <t|t> = %24.15e\n",rtmp0);
      printf("   <t|r> = %24.15e%24.15e\n",ctmp0.x,ctmp0.y);
      printf("     omg = %24.15e%24.15e\n",omg.x,omg.y);
    }
#endif

    //
    // x = x + omg v
    // r = r - omg t
    //
#ifdef _PREC_
    accum_add_mult(yded, omgd,ved);
#else
    accum_add_mult(yded, omgd,red);
#endif
    accum_add_mult( red,nomgd,ted);
    solv_flop_count += QACCUM_ADD_CMULT*2;

    rtmp0 = abs2(red);
#ifdef _DEBUG_
    err = sqrt(rtmp0);
    if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    solv_flop_count += QABS2;

    err = sqrt(rtmp0);
    maxerr = MAX(maxerr,err);
    if ( err < minerr) {
      assign(xmind,yded);
      minerr = err;
    }

    if (rtmp0 < tol2) break;

    ////////////////////////////////
    // calc true residual
    ////////////////////////////////
    if ( err < delta*maxerr ) {
      if (is_clover) {
        assign_mult_eoprec_wd_eo(mkappa2d,ved,yded,ud,fclinvd);
      } else {
        assign_mult_eoprec_wd_eo(mkappa2d,ved,yded,ud);
      }

      // red  = yed - ved = yed - Dee yded
      assign_sub(red,yed,ved);
      rtmp0 = abs2(red);
      rtmp0 = sqrt(rtmp0);
#ifdef _DEBUG_
      if ( 0 == _nodeid_ ) printf("%%%5d ERR:%15.7E\n",*iter,rtmp0);
      fflush(stdout);
#endif
      if ( rtmp0 > 2.0f*err ) break;
      err = rtmp0;
      maxerr = err;
    }

    //
    // rho1 = <rt|r>
    //
    rho1 = prod(rted,red);
    solv_flop_count += QPROD;

#ifdef _DEBUG_
    if ( 0 == _nodeid_ ) {
      printf("   rho1 = %24.15e%24.15e\n",rho1.x,rho1.y);
    }
#endif

    //
    // bet = (alp/omg)*(rho1/rho0)
    // rho0 = rho1;
    //
    ctmp0 = assign_div(alp,omg);
    ctmp1 = assign_div(rho1,rho0);
      bet = assign_mul(ctmp0,ctmp1);
    if ( sqrt(bet.x*bet.x + bet.y*bet.y) < tol_bk) {
      ierr_flag = ERROR_BREAKDOWN;
      break;
    }

    assign(betd,bet);
     rho0 = rho1;

#ifdef _DEBUG_
    if ( 0 == _nodeid_ ) {
      printf("   alp/omg = %24.15e%24.15e\n",ctmp0.x,ctmp0.y);
      printf(" rho1/rho0 = %24.15e%24.15e\n",ctmp1.x,ctmp1.y);
      printf("       bet = %24.15e%24.15e\n",  bet.x,  bet.y);
    }
#endif

    //
    // p = r + bet ( p - omg q)
    //
    accum_add_mult(ped,nomgd,qed);
    accum_mult_add(ped, betd,red);
    solv_flop_count += QACCUM_ADD_CMULT + QACCUM_CMULT_ADD;

  } // end while

  //======================================
  // check iteration counts (max iter)
  // if iter exceeds max iteration
  // flip the sign.
  //======================================
  if (*iter >= niter) { *iter = -(*iter); }


  //======================================
  // receive result from device to host
  //======================================
  startTimer(yrecv_timer);
  if ( NOERROR == ierr_flag) {
    assign(yde,yded);
  } else {
    assign(yde,xmind);
  }
  stopTimer(yrecv_timer);

#ifdef _DEBUG_
  //======================================
  // Check residual
  //
  // ved = Dee yded
  //
  if ( NOERROR == ierr_flag) {
    if (is_clover) {
      assign_mult_eoprec_wd_eo(mkappa2d,ved,yded,ud,fclinvd);
    } else {
      assign_mult_eoprec_wd_eo(mkappa2d,ved,yded,ud);
    }
  } else {
    if (is_clover) {
      assign_mult_eoprec_wd_eo(mkappa2d,ved,xmind,ud,fclinvd);
    } else {
      assign_mult_eoprec_wd_eo(mkappa2d,ved,xmind,ud);
    }
  }
  //
  // red = yed - ved
  //
  assign_sub(red,yed,ved);

  rtmp0 = abs2(red);

  err = sqrt(rtmp0);
  if ( 0 == _nodeid_ ) printf("%%%5d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
  //======================================
#endif

  //======================================
  // deallocate device memory
  //======================================
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {

    switch_gpu_(&igx);

    CUDA_SAFE_CALL(cudaFree(    alpd.g[igx]));
    CUDA_SAFE_CALL(cudaFree(   nalpd.g[igx]));
    CUDA_SAFE_CALL(cudaFree(    betd.g[igx]));
    CUDA_SAFE_CALL(cudaFree(    omgd.g[igx]));
    CUDA_SAFE_CALL(cudaFree(   nomgd.g[igx]));

    CUDA_SAFE_CALL(cudaFree(mkappa2d.g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*yed).g[igx]));
    CUDA_SAFE_CALL(cudaFree( (*yded).g[igx]));

    CUDA_SAFE_CALL(cudaFree(  (*red).g[igx]));
    CUDA_SAFE_CALL(cudaFree( (*rted).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*ped).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*qed).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*ted).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*ved).g[igx]));
    CUDA_SAFE_CALL(cudaFree((*xmind).g[igx]));
  }

  _mm_free( yed);
  _mm_free(yded);
  _mm_free( red);
  _mm_free(rted);
  _mm_free( ped);
  _mm_free( qed);
  _mm_free( ted);
  _mm_free( ved);
  _mm_free(xmind);

#ifdef _DEBUG_
  fflush(stdout);
#endif
  stopTimer(solv_timer);

}
