extern "C" void
cuda_assign_inv_mult_eoprec_wd_cg_(double   *kappa,
                                   double   *tol,
                                   int      *iter,
                                   f_wqf_eo *yde,
                                   f_wqf_eo *ye) {
//
// MAIN GPU CGNE Solver for
//  v
//  Dee yde = ye
//
//  with even/odd site preconditioned Wilson-Clover-Dirac opeartor
//    v
//    Dee = 1 - kappa^2 Fee Meo Foo Moe
//

  //===========================================================
  // host variables
  //===========================================================
  const int niter = 10000;
  int ieo;
  float err;
  float mkappa2 = -(float)((*kappa)*(*kappa));
  float    tol2 =  (float)((*tol)*(*tol));
  float alp, nalp, bet, rho0, rho1, rtmp0;

  //===========================================================
  // device variables
  //===========================================================
  float_gpu  mkappa2d;
  float_gpu  alpd, nalpd, betd;
  cu_wqf_eo *yed,*yded,*ped,*qed,*red,*sed;

#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    switch_gpu_(&igx);
#if CUDA_VERSION >= 4000
    CUDA_SAFE_CALL(cudaDeviceSynchronize());
#else
    CUDA_SAFE_CALL(cudaThreadSynchronize());
#endif
  }

  startTimer(solv_timer);

#ifdef _DEBUG_
  fflush(stdout);
  if (is_clover) {
    printf("%%S_CG_CLOVER(MPI+GPU-PARALLEL)\n");
  } else {
    printf("%%S_CG_WILSON(MPI+GPU-PARALLEL)\n");
  }
  fflush(stdout);
#endif

   ieo = (*ye).ieo;

  yded = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   yed = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   ped = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   qed = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   red = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   sed = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);

  (*yded).ieo = ieo;
   (*yed).ieo = ieo;
   (*ped).ieo = ieo;
   (*qed).ieo = ieo;
   (*red).ieo = ieo;
   (*sed).ieo = ieo;

  (*yded).ipeo = _ipeo_;
   (*yed).ipeo = _ipeo_;
   (*ped).ipeo = _ipeo_;
   (*qed).ipeo = _ipeo_;
   (*red).ipeo = _ipeo_;
   (*sed).ipeo = _ipeo_;

  //======================================
  // allocate device memory
  //======================================
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    switch_gpu_(&igx);

    CUDA_SAFE_CALL(cudaMalloc((void**)    &(alpd.g[igx]),sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void**)   &(nalpd.g[igx]),sizeof(float)));
    CUDA_SAFE_CALL(cudaMalloc((void**)    &(betd.g[igx]),sizeof(float)));

    CUDA_SAFE_CALL(cudaMalloc((void**)&(mkappa2d.g[igx]),sizeof(float)));

    CUDA_SAFE_CALL(cudaMalloc((void**) &((*yded).g[igx]),sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)  &((*yed).g[igx]),sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)  &((*ped).g[igx]),sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)  &((*qed).g[igx]),sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)  &((*red).g[igx]),sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)  &((*sed).g[igx]),sizeof(cu_wqf_eo_blk)));

    new_q((*yded).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*yed).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*ped).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*qed).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*red).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
    new_q( (*sed).g[igx],ipeo_d.g[igx],ieo_d[ieo].g[igx],igx_d.g[igx],ipsite_d.g[igx]);
  }

#ifdef _DEBUG_
  printf("%% cudaMalloc OK.\n");
#endif

//======================================
// Send data from host to device
//======================================
  startTimer(ysend_timer);
  assign(mkappa2d,mkappa2);
  assign(yed,ye);
  stopTimer(ysend_timer);

///////////////////////////////////////////////////////////////////////////////
  //
  // set x_0 = 0 (yde=0)
  //
  clear(yded,ieo);

  //
  // r  = b ( = y : rhs of equation )
  // s  = Dee^+ r
  // p  = s
  //
  assign( red,yed);
  if (is_clover) {

    assign_mult_eoprec_dag_wd_eo(mkappa2d,sed,red,ud,fclinvd);

  } else {

    assign_mult_eoprec_dag_wd_eo(mkappa2d,sed,red,ud);

  }
  solv_flop_count += MULT_MB_PRE;
  *iter = *iter + 1;
  assign( ped,sed);

  //
  // rho0 = |se|^2
  //
  rho0 = abs2(sed);
  solv_flop_count += QABS2;

#ifdef _DEBUG_
  rtmp0 = abs2(red);
  solv_flop_count += QABS2;
  err = sqrt(rtmp0);
  if ( 0 == _nodeid_ ) printf("%% S_TOL=%24.15e\n",sqrt(tol2));
  if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
#endif

  while (*iter < niter) {

    //
    // q = Dee p 
    //
    if (is_clover) {

      assign_mult_eoprec_wd_eo(mkappa2d,qed,ped,ud,fclinvd);

    } else {

      assign_mult_eoprec_wd_eo(mkappa2d,qed,ped,ud);

    }
    solv_flop_count += MULT_MB_PRE;
    *iter = *iter + 1;

    //
    // alp = rho0/<q|q>
    //
    rtmp0 = abs2(qed);
     alp = rho0/rtmp0;
    nalp = -alp;
    assign( alpd, alp);
    assign(nalpd,nalp);
    solv_flop_count += QABS2;

    //
    // x = x + alp p
    // r = r - alp q
    //
    accum_add_mult(yded, alpd,ped);
    accum_add_mult( red,nalpd,qed);
    solv_flop_count += QACCUM_ADD_RMULT*2;

    rtmp0 = abs2(red);
    solv_flop_count += QABS2;
#ifdef _DEBUG_
    err = sqrt(rtmp0);
    if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    if (rtmp0 < tol2) break;

    //
    // s = Dee^+ r
    //
    if (is_clover) {

      assign_mult_eoprec_dag_wd_eo(mkappa2d,sed,red,ud,fclinvd);

    } else {

      assign_mult_eoprec_dag_wd_eo(mkappa2d,sed,red,ud);

    }
    solv_flop_count += MULT_MB_PRE;
    *iter = *iter + 1;

    //
    // rho1 = |s|^2
    // bet  = rho1/rho0
    // rho0 = rho1
    //
    rho1 = abs2(sed);
    solv_flop_count += QABS2;
    bet  = rho1/rho0;
    rho0 = rho1;
    assign( betd, bet);

    // p = s + bet p
    accum_mult_add(ped, betd,sed);
    solv_flop_count += QACCUM_RMULT_ADD;

  } // end while

  //======================================
  // check iteration counts (max iter)
  // if iter exceeds max iteration
  // flip the sign.
  //======================================
  if (*iter >= niter) {
    rtmp0 = abs2(red);
    solv_flop_count += QABS2;
    err = sqrt(rtmp0);
    if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
    if ( 0 == _nodeid_ ) printf("@%s %d : Maximum iteration limit was reached. Increase niter=%d in %s.\n",__FILE__,__LINE__,niter,__FILE__);
    fflush(stdout);
    *iter = -(*iter);
  }

  //======================================
  // receive result from device to host
  //======================================
  startTimer(yrecv_timer);
  assign(yde,yded);
  stopTimer(yrecv_timer);

#ifdef _DEBUG_
  //
  // Check residual
  //
  // sed = Dee yded
  //
  if (is_clover) {

    assign_mult_eoprec_wd_eo(mkappa2d,sed,yded,ud,fclinvd);

  } else {

    assign_mult_eoprec_wd_eo(mkappa2d,sed,yded,ud);

  }
  //
  // red = yed - sed
  //
  assign_sub(red,yed,sed);

  rtmp0 = abs2(red);
  err = sqrt(rtmp0);
  if ( 0 == _nodeid_ ) printf("%%%5d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
#endif

//======================================
// deallocate device memory
//======================================
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {

    switch_gpu_(&igx);

    CUDA_SAFE_CALL(cudaFree(    alpd.g[igx]));
    CUDA_SAFE_CALL(cudaFree(   nalpd.g[igx]));
    CUDA_SAFE_CALL(cudaFree(    betd.g[igx]));

    CUDA_SAFE_CALL(cudaFree(mkappa2d.g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*yed).g[igx]));
    CUDA_SAFE_CALL(cudaFree( (*yded).g[igx]));

    CUDA_SAFE_CALL(cudaFree(  (*ped).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*qed).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*red).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*sed).g[igx]));

  }

  _mm_free(yded);
  _mm_free( yed);
  _mm_free( ped);
  _mm_free( qed);
  _mm_free( red);
  _mm_free( sed);

#ifdef _DEBUG_
  fflush(stdout);
#endif
  stopTimer(solv_timer);
}
