static void 
assign_mult_eoprec_wd_eo(const float_gpu &mkappa2, 
                               cu_wqf_eo *dye, 
                               cu_wqf_eo *ye, 
                            const cu_gvf *u, 
                           const cu_clvf *fclinv = NULL) {
//
//                                                             v
// Multiply even/odd-site preconditioned Dirac operator : yd = Dee y,
//       v
// where Dee = 1 - kappa^2 Fee Meo Foo Moe
//
//  mkappa2 : hopping parameter (= -kappa^2)
//      dye : even/odd site fermion (output)
//       ye : even/odd site fermion (input)
//        u : gauge link
//   fclinv : inverse clover term
//

  cu_wqf_eo *Myo;
  int ieo = (*ye).ieo;
  int ioe = 1-ieo;

  copy_boundary(ye);

  startTimer(mult_timer);
  Myo = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
  (*Myo).ipeo = _ipeo_;
  (*Myo).ieo  = ioe;

  //
  //  Myo = Moe ye, or  Foo \ Moe ye
  //
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {

    switch_gpu_(&igx);
    CUDA_SAFE_CALL(cudaMalloc((void**)&((*Myo).g[igx]),sizeof(cu_wqf_eo_blk)));
    new_q((*Myo).g[igx],ipeo_d.g[igx],ieo_d[ioe].g[igx],igx_d.g[igx],ipsite_d.g[igx]);

#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_yori,       (*ye).g[igx],sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_ueri,(*u).eo[ioe].g[igx],sizeof(cu_gvf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_uori,(*u).eo[ieo].g[igx],sizeof(cu_gvf_eo_blk)));
#endif

    if (is_clover) {
      assign_mult_hop_in_kernel<<<_NBLOCKS,_NTHREADS>>>((*Myo).g[igx],phys_blk_d.g[igx],
                                                         (*ye).g[igx],
                                                  (*u).eo[ioe].g[igx], 
                                                  (*u).eo[ieo].g[igx],
                                             (*fclinv).eo[ioe].g[igx]);
    } else {
      assign_mult_hop_in_kernel<<<_NBLOCKS,_NTHREADS>>>((*Myo).g[igx],phys_blk_d.g[igx],
                                                         (*ye).g[igx],
                                                  (*u).eo[ioe].g[igx],
                                                  (*u).eo[ieo].g[igx]);
    }
    CUT_CHECK_ERROR("cumult_wd_in_kernel (all) fails.");

#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_yori));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_ueri));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_uori));
#endif

  }
  stopTimer(mult_timer);

  copy_boundary(Myo);

  //
  //  dye = Meo Myo, or  Fee \ Meo Myo
  //
  startTimer(mult_timer);
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    switch_gpu_(&igx);

#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_yori,      (*Myo).g[igx],sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_ueri,(*u).eo[ieo].g[igx],sizeof(cu_gvf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_uori,(*u).eo[ioe].g[igx],sizeof(cu_gvf_eo_blk)));
#endif

    if (is_clover) {
      assign_mult_hop_in_kernel<<<_NBLOCKS,_NTHREADS>>>((*dye).g[igx],phys_blk_d.g[igx],
                                                        (*Myo).g[igx],
                                                  (*u).eo[ieo].g[igx],
                                                  (*u).eo[ioe].g[igx],(*fclinv).eo[ieo].g[igx]);
    } else {
      assign_mult_hop_in_kernel<<<_NBLOCKS,_NTHREADS>>>((*dye).g[igx],phys_blk_d.g[igx],
                                                        (*Myo).g[igx],
                                                  (*u).eo[ieo].g[igx],
                                                  (*u).eo[ioe].g[igx]);
    }
    CUT_CHECK_ERROR("cumult_wd_in_kernel (2all) fails.");

#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_yori));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_ueri));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_uori));
#endif

    CUDA_SAFE_CALL(cudaFree((*Myo).g[igx]));
  }
  _mm_free(Myo);
  stopTimer(mult_timer);

  if (is_clover) {
    mult_flop_count += 2*MULT_HOP + 2*MULT_CLOVER;
  } else {
    mult_flop_count += 2*MULT_HOP;
  }
  _multiter_ += 2;

  //
  // dye = - kappa^2 dye + ye
  //
  accum_mult_add(dye, mkappa2, ye);

}

static void 
assign_mult_eoprec_dag_wd_eo(const float_gpu &mkappa2, 
                               cu_wqf_eo *dye, 
                               cu_wqf_eo *ye, 
                           const cu_gvf  *u, 
                           const cu_clvf *fclinv = NULL) {
//
//                                                                      v
// Multiply even/odd-site daggered-preconditioned Dirac operator : yd = Dee^+ y,
//       v
// where Dee^+ = 1 - kappa^2 gamma5 Meo Foo Moe Fee gamma5
//
//  mkappa2 : hopping parameter (= -kappa^2)
//      dye : even/odd site fermion (output)
//       ye : even/odd site fermion (input)
//        u : gauge link
//   fclinv : inverse clover term
//

  cu_wqf_eo *Myo;
  int ieo = (*ye).ieo;
  int ioe = 1-ieo;

  startTimer(mult_timer);

  if (is_clover) {

    //
    // yde = gamma5 Fee ye
    //
    assign_mult_clover_term(dye,ye,fclinv);
    accum_mult_gamma5(dye);

  } else {

    //
    // yde = gamma5 ye
    //
    assign_mult_gamma5(dye,ye);

  }

  stopTimer(mult_timer);

  copy_boundary(dye);

  startTimer(mult_timer);

  Myo = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
  (*Myo).ipeo = _ipeo_;
  (*Myo).ieo  = ioe;

  //
  //  Myo = Moe yde, or  Foo \ Moe yde
  //
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {

    switch_gpu_(&igx);

    CUDA_SAFE_CALL(cudaMalloc((void**)&((*Myo).g[igx]),sizeof(cu_wqf_eo_blk)));
    new_q((*Myo).g[igx],ipeo_d.g[igx],ieo_d[ioe].g[igx],igx_d.g[igx],ipsite_d.g[igx]);

#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_yori,      (*dye).g[igx],sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_ueri,(*u).eo[ioe].g[igx],sizeof(cu_gvf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_uori,(*u).eo[ieo].g[igx],sizeof(cu_gvf_eo_blk)));
#endif

    if (is_clover) {
      assign_mult_hop_in_kernel<<<_NBLOCKS,_NTHREADS>>>((*Myo).g[igx],phys_blk_d.g[igx],
                                                        (*dye).g[igx],
                                                  (*u).eo[ioe].g[igx], 
                                                  (*u).eo[ieo].g[igx],
                                             (*fclinv).eo[ioe].g[igx]);
    } else {
      assign_mult_hop_in_kernel<<<_NBLOCKS,_NTHREADS>>>((*Myo).g[igx],phys_blk_d.g[igx],
                                                        (*dye).g[igx],
                                                  (*u).eo[ioe].g[igx],
                                                  (*u).eo[ieo].g[igx]);
    }
    CUT_CHECK_ERROR("cumult_wd_in_kernel (all) fails.");

#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_yori));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_ueri));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_uori));
#endif

  }
  stopTimer(mult_timer);

  copy_boundary(Myo);

  //
  // dye = Meo Myo
  //
  startTimer(mult_timer);
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {

    switch_gpu_(&igx);

#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_yori,      (*Myo).g[igx],sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_ueri,(*u).eo[ieo].g[igx],sizeof(cu_gvf_eo_blk)));
    CUDA_SAFE_CALL(cudaBindTexture(0,tex_uori,(*u).eo[ioe].g[igx],sizeof(cu_gvf_eo_blk)));
#endif

    assign_mult_hop_in_kernel<<<_NBLOCKS,_NTHREADS>>>((*dye).g[igx],phys_blk_d.g[igx],
                                                      (*Myo).g[igx],
                                                (*u).eo[ieo].g[igx],
                                                (*u).eo[ioe].g[igx]);

    CUT_CHECK_ERROR("cumult_wd_in_kernel (2all) fails.");

#ifdef _USE_TEXFETCH_
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_yori));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_ueri));
    CUDA_SAFE_CALL(cudaUnbindTexture(tex_uori));
#endif

    CUDA_SAFE_CALL(cudaFree((*Myo).g[igx]));
  }
  _mm_free(Myo);

  //
  // dye = gamma5 dye
  //
  accum_mult_gamma5(dye);

  stopTimer(mult_timer);

  if (is_clover) {
    mult_flop_count += 2*MULT_HOP + 2*MULT_CLOVER;
  } else {
    mult_flop_count += 2*MULT_HOP;
  }
  _multiter_ += 2;

  //
  // dye = - kappa^2 dye + ye
  //
  accum_mult_add(dye, mkappa2, ye);

}
