///////////////////////////////////////////////
// Single precision
///////////////////////////////////////////////

#define _ALIG_ __align__(16)

#define NDIM   (4)
#define COL    (3)
#define CLCL   (COL*COL)
#define SPIN   (4)
#define CLSP   (COL*SPIN)
#define CSH    (COL*SPIN/2)
#define CLSPH  ((CLSP/2+1)*(CLSP/4))

#define _FULL_BLOCK_   (_NBZ*(_NBY+_NDEPTH*2)*(_NBX+_NDEPTH*2))
#define _PHYS_BLOCK_   (_NBZ*_NBY*_NBX)
#define _IN_BLOCK_     (_NBZ*(_NBY-2)*(_NBX-2))
#define _OUT_BLOCK_    (_NBZ*(_NBY*_NBX-(_NBY-2)*(_NBX-2)))
#define _NBXYZ_Q_      (_NBZ*(_NBY+_NDEPTH*2)*(_NBX+_NDEPTH*2))
#define _NBXYZ_G_      (_NBZ*(_NBY+_NDEPTH*2)*(_NBX+_NDEPTH*2))
#define _NBXYZ_C_      (_NBZ*(_NBY+_NDEPTH*2)*(_NBX+_NDEPTH*2))
#define _NTHZT_        (_NTHZ*_NTHT)

#define  thd_t         ( threadIdx.x % _NTHT)
#define  thd_z         ((threadIdx.x / _NTHT) % _NTHZ)
#define  blk_z(iblk_)  (  (iblk_) % _NBZ)
#define  blk_y(iblk_)  ((((iblk_) / _NBZ)% (_NBY+_NDEPTH*2))+1-_NDEPTH-1)
#define  blk_x(iblk_)  ((((iblk_) / _NBZ / (_NBY+_NDEPTH*2) ) % (_NBX+_NDEPTH*2))+1-_NDEPTH-1)

#define  blk_idx(ibz,iby,ibx)  (((ibz)-1)+((iby)-1+_NDEPTH)*_NBZ+((ibx)-1+_NDEPTH)*_NBZ*(_NBY+_NDEPTH*2))

////////////////////////////
// Texture fetching
// indexing macro
////////////////////////////
#define __loadu_tex(turi,ithd,ic,mu,iblk)  tex1Dfetch(turi,(ithd)+_NTHREADS*(ic )+_NTHREADS*COL*(mu)+_NTHREADS*COL*NDIM*(iblk))
#define __loady_tex(tyri,ithd,ics,iblk)    tex1Dfetch(tyri,(ithd)+_NTHREADS*(ics)+_NTHREADS*CSH*(iblk))

//////////////////////
// my complex type
//////////////////////
typedef float2 cuComplex;

//////////////////////
// device data
//////////////////////
typedef struct _ALIG_ {
  int *g[_NGX];
  int  cpu_val;
} int_gpu;

typedef struct _ALIG_ {
  int4 *g[_NGX];
  int4  cpu_val;
} int4_gpu;

typedef struct _ALIG_ {
  float *g[_NGX];
  float  cpu_val;
} float_gpu;

typedef struct _ALIG_ {
  cuComplex *g[_NGX];
  cuComplex  cpu_val;
} cuComplex_gpu;

/////////////////////////
// float4 thread vector
/////////////////////////
typedef struct _ALIG_ {
  float4 f[_NTHZT_];
} cufloat4_thread;

////////////////
// quark field
////////////////
typedef struct _ALIG_ {
  cufloat4_thread thd[CSH];
} cu_wqf_eo_thd;

typedef struct _ALIG_ {
  cu_wqf_eo_thd blk[_NBXYZ_Q_];
  int  igx;
  int  ieo;
  int  ipeo;
  int  dummy2;
  int4 ipsite;
} cu_wqf_eo_blk;

typedef struct _ALIG_ {
  cu_wqf_eo_blk *g[_NGX];
  int ieo;
  int ipeo;
  int dummy2;
  int dummy3;
} cu_wqf_eo;

typedef struct _ALIG_ {
  cu_wqf_eo eo[2];
  int ipeo;
  int dummy;
} cu_wqf;

typedef struct _ALIG_ {
  cu_wqf_eo_blk g[_NGX];
  int ieo;
  int ipeo;
  int dummy2;
  int dummy3;
} f_wqf_eo;

typedef struct _ALIG_ {
  f_wqf_eo eo[2];
  int ipeo;
  int dummy;
} f_wqf;

////////////////
// gauge field
////////////////
typedef struct _ALIG_ {
  cufloat4_thread thd[NDIM][COL];
} cu_gvf_eo_thd;

typedef struct _ALIG_ {
  cu_gvf_eo_thd blk[_NBXYZ_G_];
  int  igx;
  int  ieo;
  int  ipeo;
  int  dummy2;
  int4 ipsite;
} cu_gvf_eo_blk;

typedef struct _ALIG_ {
  cu_gvf_eo_blk *g[_NGX];
  int ieo;
  int ipeo;
  int dummy2;
  int dummy3;
} cu_gvf_eo;

typedef struct _ALIG_ {
  cu_gvf_eo eo[2];
  int ipeo;
  int dummy;
} cu_gvf;

typedef struct _ALIG_ {
  cu_gvf_eo_blk g[_NGX];
  int ieo;
  int ipeo;
  int dummy2;
  int dummy3;
} f_gvf_eo;

typedef struct _ALIG_ {
  f_gvf_eo eo[2];
  int ipeo;
  int dummy;
} f_gvf;

////////////////
// clover field
////////////////
typedef struct _ALIG_ {
  cufloat4_thread thd[CLSPH];
} cu_clvf_eo_thd;

typedef struct _ALIG_ {
  cu_clvf_eo_thd blk[_NBXYZ_C_];
  int  igx;
  int  ieo;
  int  ipeo;
  int  dummy2;
  int4 ipsite;
} cu_clvf_eo_blk;

typedef struct _ALIG_ {
  cu_clvf_eo_blk *g[_NGX];
  int ieo;
  int ipeo;
  int dummy2;
  int dummy3;
} cu_clvf_eo;

typedef struct _ALIG_ {
  cu_clvf_eo eo[2];
  int ipeo;
  int dummy;
} cu_clvf;

typedef struct _ALIG_ {
  cu_clvf_eo_blk g[_NGX];
  int ieo;
  int ipeo;
  int dummy2;
  int dummy3;
} f_clvf_eo;

typedef struct _ALIG_ {
  f_clvf_eo eo[2];
  int ipeo;
  int dummy;
} f_clvf;

