#ifndef _ASSIGN_MULT_EOPREC_WD_EO_MIC_H
#define _ASSIGN_MULT_EOPREC_WD_EO_MIC_H

extern "C" void 
#ifdef _CLOVER
assign_mult_eoprec_wd_eo_mic(const float      *kappa,
                                   mic_wqf_eo *restrict dye,
                             const mic_wqf_eo *restrict  ye,
                             const mic_gvf_soa    *u,
                             const mic_clvf   *fclinv)
#else
assign_mult_eoprec_wd_eo_mic(const float      *kappa,
                                   mic_wqf_eo *restrict dye,
                             const mic_wqf_eo *restrict  ye,
                             const mic_gvf_soa    *u) 
#endif
{
//
//                                                             v
// Multiply even/odd-site preconditioned Dirac operator : yd = Dee y,
//       v
// where Dee = 1 - kappa^2 Fee Meo Foo Moe
//
//  mkappa2 : hopping parameter (= -kappa^2)
//      dye : even/odd site fermion (output)
//       ye : even/odd site fermion (input)
//        u : gauge link
//   fclinv : inverse clover term
//

  mic_wqf_eo *Myo;
  int ieo  = (*ye).ieo;
  int ioe  = 1-ieo;

  Myo = (mic_wqf_eo *)_mm_malloc(sizeof(mic_wqf_eo),_ALIGN_SIZE);
  (*Myo).ieo  = ioe;

#ifdef _TIMING_
  double multfpc0 = solver_timer.mult_flop_count;
#endif

  //
  //  Myo = Moe ye, or  Foo \ Moe ye
  //
#ifdef _CLOVER
  assign_mult_hop_eo_mic(Myo, ye,&((*u).eo[ioe]),&((*u).eo[ieo]),&((*fclinv).eo[ioe]));
#else
  assign_mult_hop_eo_mic(Myo, ye,&((*u).eo[ioe]),&((*u).eo[ieo]));
#endif

  //
  //  dye = Meo Myo, or  Fee \ Meo Myo
  //
#ifdef _CLOVER
  assign_mult_hop_eo_mic(dye,Myo,&((*u).eo[ieo]),&((*u).eo[ioe]),&((*fclinv).eo[ieo]));
#else
  assign_mult_hop_eo_mic(dye,Myo,&((*u).eo[ieo]),&((*u).eo[ioe]));
#endif

  _mm_free(Myo);

  //
  // dye = - kappa^2 dye + ye
  //
  _ALIG_PRE_ float mkappa2 _ALIG_PST_;
  mkappa2 = - (*kappa)*(*kappa);
  accum_mult_add(dye, &mkappa2, ye);

  assert(dye->ieo == ye->ieo);

//prcorners("dye1", dye);

#ifdef _TIMING_
  solver_timer.solv_flop_count += (solver_timer.mult_flop_count - multfpc0) + (QACCUM_RMULT_ADD)*NTH*NZ*NY*NX;
#endif

}

#endif
