#include <omp.h>
#include <time.h>
#include "micTimer.h"

#define _RDTSC

//////////////////////////////////////////
// Intel Xeon Phi Coprocessor 7110P
//  1.1GHz
//////////////////////////////////////////
#if __MIC__ == 1
#define _CPUHZ_ (1100000000)
#elif __AVX__ == 1
#define _CPUHZ_ (2500000000)
#endif

static inline unsigned long rdtsc() {

  unsigned int hi, lo;

  __asm__ __volatile__ (
     "xorl %%eax, %%eax \n\t"
     "cpuid             \n\t"
     "rdtsc             \n\t"
     :"=a"(lo), "=d"(hi)
     :
     :"%ebx", "%ecx"
  );
  return ((unsigned long)hi << 32) | lo;
}

extern "C"
void __createTimer(micTimer_t &timer) {
        #pragma omp critical (create_timer)
        {
                timer.start = 0.0;
                timer.stop  = 0.0;
                timer.et    = 0.0;
        }
}

extern "C"
void __deleteTimer(micTimer_t &timer) {
        #pragma omp critical (delete_timer)
        {
                timer.start = 0.0;
                timer.stop  = 0.0;
                timer.et    = 0.0;
        }
}

extern "C"
void __resetTimer(micTimer_t &timer) {
        #pragma omp critical (reset_timer)
        {
                timer.start = 0.0;
                timer.stop  = 0.0;
                timer.et    = 0.0;
        }
}

extern "C"
void __startTimer(micTimer_t &timer) {

//        timer.start = omp_get_wtime();

#ifdef _RDTSC
  timer.start = (double) rdtsc();
#else
  struct timespec tsp;
  clock_gettime(CLOCK_MONOTONIC, &tsp);
  timer.start = (double) tsp.tv_sec + (double) tsp.tv_nsec * 1e-9;
#endif

}

extern "C"
void __stopTimer(micTimer_t &timer) {

//        timer.stop = omp_get_wtime();
//        double tmp = timer.stop - timer.start;
//        timer.et += tmp;

#ifdef _RDTSC
  timer.stop = (double) rdtsc();
#else
  struct timespec tsp;
  clock_gettime(CLOCK_MONOTONIC, &tsp);
  timer.stop = (double) tsp.tv_sec + (double) tsp.tv_nsec * 1e-9;
#endif

  double tmp = timer.stop - timer.start;
  timer.et += tmp;

}

extern "C"
double __getTimerValue(micTimer_t &timer) {
        return timer.et/_CPUHZ_;
}

