#ifndef _MIC_CONFIG_H
#define _MIC_CONFIG_H

!------------------------------------------------------------------
! if _DEBUG_  is defined, the residual history will be printed out.
! if _TIMING_ is defined, the solver flops performance will be measured.
!------------------------------------------------------------------
!#define _DEBUG_
!#define _TIMING_

!----------------------------
! choice of preconditioner
! Only RAS is available
!----------------------------
#define _USE_RAS

!------------------------------------------------------------------
! RAS (Restricted overlapping Additiv Shwarz preconditioner)
! MPI Parallel only
!  _NRAS   : number of RAS iteration 
!  _NMR    : domain MR iteration
!  _NDEPTH : depth of domain overlapping region
!------------------------------------------------------------------
#ifdef _USE_RAS
#define _NRAS   (2)
#define _NMR    (2)
#define _NDEPTH (1)
#endif


!------------------------------------------------------------------
! Reserved paramters for
! SAP (Shwarz Alternating Procedure preconditioner)
!  _NSAP    : number of SAP iteration
!  _NMR_SAP : domain MR iteration 
!------------------------------------------------------------------
!#ifdef _USE_SAP
!#define _NSAP     (0)
!#define _NMR_SAP  (0)
!#endif


!-----------------------------------------
! RAS preconditioner is prefarable
!-----------------------------------------
#if defined(_USE_RAS) && defined(_USE_SAP)
#undef _USE_SAP
#endif

#ifndef _USE_RAS
#define _NRAS   (0)
#define _NMR    (0)
#define _NDEPTH (0)
#endif

#ifndef _USE_SAP
#define _NSAP     (0)
#define _NMR_SAP  (0)
#endif

#ifdef _singlePU
#endif


!------------------------------------------------------------------
! T-size check
!------------------------------------------------------------------
#define _NTTDIVVLEN   ((_NTT)/(8))
#define _NTTMODVLEN   (_NTT)-(_NTTDIVVLEN*8)

#if _NTTMODVLEN != 0
#error NTT must be a multiple of 8. Check config.h
#endif

!------------------------------------------------------------------
! fermion boundary condition check
!------------------------------------------------------------------
#if defined(_SF)
#error SF boundary is not implemented in MICKernels. Check config.h
#endif
#if _FERMI_BOUNDARY_TYPE_T != PBC
#error Only periodic boundary is defined. FERMI_BOUNDRY_TYPE_T must be PBC. Check config.h
#endif
#if _FERMI_BOUNDARY_TYPE_Z != PBC
#error Only periodic boundary is defined. FERMI_BOUNDRY_TYPE_Z must be PBC. Check config.h
#endif
#if _FERMI_BOUNDARY_TYPE_Y != PBC
#error Only periodic boundary is defined. FERMI_BOUNDRY_TYPE_Y must be PBC. Check config.h
#endif
#if _FERMI_BOUNDARY_TYPE_X != PBC
#error Only periodic boundary is defined. FERMI_BOUNDRY_TYPE_X must be PBC. Check config.h
#endif


#endif
