#ifndef _MIC_PARAM_SIZE_H
#define _MIC_PARAM_SIZE_H

// for Windows
//#define _ALIG_PRE_ __declspec(align(_ALIGN_SIZE))
//#define _ALIG_PST_

// for Linux
#define _ALIG_PRE_
#define _ALIG_PST_  __attribute__((aligned(_ALIGN_SIZE)))

#define NDIM   (4)
#define COL    (3)
#define CLCL   (COL*COL)
#define SPIN   (4)
#define CLSP   (COL*SPIN)
#define CSH    (COL*SPIN/2)
#define CLSPH  ((CLSP/2+1)*(CLSP/4))

//////////////////////
// time vector length
//////////////////////

#define VLEN (4)
#define _ALIGN_SIZE (64)


//////////////////////
// half spinor type
//////////////////////
#define __mic_half_spinor_t    __m512


#define NTT     (_NTT)
#define NT      (_NTT)
#define NX      (_NX)
#define NY      (_NY)
#define NZ      (_NZ)
#define NTH     (_NTT/2)
#define NTHT    (NTH/VLEN)
#define NTHTZYX     (NTHT*NZ*NY*NX)

/////////////////////////////////////////////////////////
// SAP even/odd domain decomposition in Z-direction
// Two domain blocks in a node.
/////////////////////////////////////////////////////////
#define NBZ        (_NZ/2)
#define NTHTBZYX   (NTHT*NBZ*NY*NX)

///////////////////////////////////////
// RAS overlap region depth
///////////////////////////////////////
#if _NDIMZ != 1
#  define NDEPTHZ  (_NDEPTH)
#else
#  define NDEPTHZ  (0)
#endif
#if _NDIMY != 1
#  define NDEPTHY  (_NDEPTH)
#else
#  define NDEPTHY  (0)
#endif
#if _NDIMX != 1
#  define NDEPTHX  (_NDEPTH)
#else
#  define NDEPTHX  (0)
#endif

#define NEX   (NX+NDEPTHX*2)
#define NEY   (NY+NDEPTHY*2)
#define NEZ   (NZ+NDEPTHZ*2)

#define NTHTZYX_RAS  (NTHT*NEZ*NEY*NEX)

///////////////////////////////////////////////////
// RAS site accessor without overlapped region
///////////////////////////////////////////////////
static inline int get_site(int itht,int iz,int iy,int ix) {
  return (itht + NTHT*(iz + NZ*(iy + NY*ix)));
}

///////////////////////////////////////////////////
// RAS site accessor with overlapped region
///////////////////////////////////////////////////
static inline int get_ras_site(int itht,int iz,int iy,int ix) {
  return (itht + NTHT*(iz + NEZ*(iy + NEY*ix)));
}

///////////////////////////////////////////////////
// SAP site accessor without overlapped region
// in a domain with a site color.
///////////////////////////////////////////////////
static inline int get_sap_eo_deo_site(int itht, int ibz, int iy, int ix) {
  return (itht + NTHT*(ibz + NBZ*(iy + NY*ix)));
}

/////////////////
// nodeinfo
/////////////////
#if 0
static int mult_iter;
static int mic_nodeid_x;
static int mic_nodeid_y;
static int mic_nodeid_z;
static int mic_nodeid_x_up;
static int mic_nodeid_y_up;
static int mic_nodeid_z_up;
static int mic_nodeid_x_dn;
static int mic_nodeid_y_dn;
static int mic_nodeid_z_dn;
static int mic_ipdeo;
static int mic_nodeid;
static int mic_ipeo;
#endif
#define _nodeid_ (mic_nodeid)

///////////////////////////////////////////////////

////////////////
// quark fields
////////////////
typedef struct _ALIG_PRE_ {
  __mic_half_spinor_t y[COL*2];
} mic_su3_spinor _ALIG_PST_;

// for RAS without overlap region
typedef struct _ALIG_PRE_ {
  mic_su3_spinor s[NTHTZYX];
  int ieo;
  int idummy[15];
} mic_wqf_eo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_wqf_eo eo[2];
} mic_wqf _ALIG_PST_;

// for RAS with overlap region
typedef struct _ALIG_PRE_ {
  mic_su3_spinor s[NTHTZYX_RAS];
  int ieo;
  int idummy[15];
} mic_ras_wqf_eo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_ras_wqf_eo eo[2];
} mic_ras_wqf _ALIG_PST_;

// for SAP 
typedef struct _ALIG_PRE_ {
  mic_su3_spinor s[NTHTBZYX];
  int ieo;
  int ideo;
  int idummy[14];
} mic_sap_wqf_eo_deo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_sap_wqf_eo_deo eo[2];
  int ideo;
  int idummy[15];
} mic_sap_wqf_deo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_sap_wqf_deo deo[2];
} mic_sap_wqf _ALIG_PST_;

////////////////
// gauge field
////////////////
typedef struct _ALIG_PRE_ {
  __mic_half_spinor_t u[COL];
} mic_su3mat _ALIG_PST_;

// for RAS without overlap region
typedef struct _ALIG_PRE_ {
  mic_su3mat s[NTHTZYX];
} mic_gsf_eo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_gsf_eo  v[NDIM];
  int ieo;
  int idummy[15];
} mic_gvf_eo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_gvf_eo eo[2];
} mic_gvf _ALIG_PST_;


// SOA change
typedef struct _ALIG_PRE_ {
  mic_su3mat v[NDIM];
} mic_gsf_eo_soa _ALIG_PST_;
typedef struct _ALIG_PRE_ {
  mic_gsf_eo_soa s[NTHTZYX];
  int ieo;
  int idummy[15];
} mic_gvf_eo_soa _ALIG_PST_;
typedef struct _ALIG_PRE_ {
  mic_gvf_eo_soa eo[2];
} mic_gvf_soa _ALIG_PST_;

// for RAS with overlap region
typedef struct _ALIG_PRE_ {
  mic_su3mat s[NTHTZYX_RAS];
} mic_ras_gsf_eo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_ras_gsf_eo  v[NDIM];
  int ieo;
  int idummy[15];
} mic_ras_gvf_eo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_ras_gvf_eo eo[2];
} mic_ras_gvf _ALIG_PST_;

// SOA
typedef struct _ALIG_PRE_ {
  mic_su3mat v[NDIM];
} mic_ras_gsf_eo_soa _ALIG_PST_;
typedef struct _ALIG_PRE_ {
  mic_ras_gsf_eo_soa  s[NTHTZYX_RAS];
  int ieo;
  int idummy[15];
} mic_ras_gvf_eo_soa _ALIG_PST_;
typedef struct _ALIG_PRE_ {
  mic_ras_gvf_eo_soa eo[2];
} mic_ras_gvf_soa _ALIG_PST_;

// for SAP 
typedef struct _ALIG_PRE_ {
  mic_su3mat s[NTHTBZYX];
} mic_sap_gsf_eo_deo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_sap_gsf_eo_deo v[NDIM];
  int ieo;
  int ideo;
  int idummy[14];
} mic_sap_gvf_eo_deo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_sap_gvf_eo_deo eo[2];
  int ideo;
  int idummy[15];
} mic_sap_gvf_deo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_sap_gvf_deo deo[2];
} mic_sap_gvf _ALIG_PST_;

/////////////////
// clover field
/////////////////
typedef struct _ALIG_PRE_ {
  __mic_half_spinor_t f[COL*COL];
} mic_half_clover _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_half_clover c[2];
} mic_clover _ALIG_PST_;

// for RAS without overlap region
typedef struct _ALIG_PRE_ {
  mic_clover s[NTHTZYX];
  int ieo;
  int idummy[15];
} mic_clvf_eo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_clvf_eo eo[2];
} mic_clvf _ALIG_PST_;

// for RAS with overlap region
typedef struct _ALIG_PRE_ {
  mic_clover s[NTHTZYX_RAS];
  int ieo;
  int idummy[15];
} mic_ras_clvf_eo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_ras_clvf_eo eo[2];
} mic_ras_clvf _ALIG_PST_;

// for SAP
typedef struct _ALIG_PRE_ {
  mic_clover s[NTHTBZYX];
  int ieo;
  int ideo;
  int idummy[14];
} mic_sap_clvf_eo_deo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_sap_clvf_eo_deo eo[2];
  int ideo;
  int idummy[15];
} mic_sap_clvf_deo _ALIG_PST_;

typedef struct _ALIG_PRE_ {
  mic_sap_clvf_deo deo[2];
} mic_sap_clvf _ALIG_PST_;


#endif
