#ifndef _MIC_RAS_INIT_H
#define _MIC_RAS_INIT_H

#ifndef _singlePU
///////////////////////////////////////////////////////////////
// GHOST SITE BUFFEFR SIZES for Half spinor (for hopping mult)
///////////////////////////////////////////////////////////////
#if _NDIMX != 1
#define NRAS_GHOST_X  (NTHT*NZ*NY*COL)
#else
#define NRAS_GHOST_X  (0)
#endif
#if _NDIMY != 1
#define NRAS_GHOST_Y  (NTHT*NZ*NX*COL)
#else
#define NRAS_GHOST_Y  (0)
#endif
#if _NDIMZ != 1
#define NRAS_GHOST_Z  (NTHT*NY*NX*COL)
#else
#define NRAS_GHOST_Z  (0)
#endif

//////////////////////////////////////////////////////////
// SCIF(reverse-offloded to the MPI on the host) BUFF
//
// First set is for assign_mult_hop_eo_mic
// Second set is for copy_boundary_ras_wqf_eo
//////////////////////////////////////////////////////////
#if _NDIMX != 1

static __mic_half_spinor_t *sbuff_x_up;//[NRAS_GHOST_X] _ALIG_PST_;
static __mic_half_spinor_t *sbuff_x_dn;//[NRAS_GHOST_X] _ALIG_PST_;
static __mic_half_spinor_t *rbuff_x_up;//[NRAS_GHOST_X] _ALIG_PST_;
static __mic_half_spinor_t *rbuff_x_dn;//[NRAS_GHOST_X] _ALIG_PST_;
static off_t bs_x_up, bs_x_dn, br_x_up, br_x_dn;

static const int bsize_x = sizeof(mic_su3_spinor)*NTHT*(NZ+2*NDEPTHZ)*(NY+2*NDEPTHY);
static __mic_half_spinor_t  *sxbuff_up, *sxbuff_dn;
static __mic_half_spinor_t  *rxbuff_up, *rxbuff_dn;
static off_t bsxbuff_up, bsxbuff_dn;
static off_t brxbuff_up, brxbuff_dn;

#endif

#if _NDIMY != 1

static __mic_half_spinor_t *sbuff_y_up;//[NRAS_GHOST_Y] _ALIG_PST_;
static __mic_half_spinor_t *sbuff_y_dn;//[NRAS_GHOST_Y] _ALIG_PST_;
static __mic_half_spinor_t *rbuff_y_up;//[NRAS_GHOST_Y] _ALIG_PST_;
static __mic_half_spinor_t *rbuff_y_dn;//[NRAS_GHOST_Y] _ALIG_PST_;
static off_t bs_y_up, bs_y_dn, br_y_up, br_y_dn;

static const int bsize_y = sizeof(mic_su3_spinor)*NTHT*(NZ+2*NDEPTHZ)*NX;
static __mic_half_spinor_t  *sybuff_up, *sybuff_dn;
static __mic_half_spinor_t  *rybuff_up, *rybuff_dn;
static off_t bsybuff_up, bsybuff_dn;
static off_t brybuff_up, brybuff_dn;

#endif

#if _NDIMZ != 1

static __mic_half_spinor_t *sbuff_z_up;//[NRAS_GHOST_Z] _ALIG_PST_;
static __mic_half_spinor_t *sbuff_z_dn;//[NRAS_GHOST_Z] _ALIG_PST_;
static __mic_half_spinor_t *rbuff_z_up;//[NRAS_GHOST_Z] _ALIG_PST_;
static __mic_half_spinor_t *rbuff_z_dn;//[NRAS_GHOST_Z] _ALIG_PST_;
static off_t bs_z_up, bs_z_dn, br_z_up, br_z_dn;

static const int bsize_z = sizeof(mic_su3_spinor)*NTHT*NX*NY;
static __mic_half_spinor_t  *szbuff_up, *szbuff_dn;
static __mic_half_spinor_t  *rzbuff_up, *rzbuff_dn;
static off_t bszbuff_up, bszbuff_dn;
static off_t brzbuff_up, brzbuff_dn;

#endif

#endif

///////////////////////////////////////////////////////////////
// Keep RAS domain extended fields (global)
//                  accessed by the RAS preconditioner
// The memory must be allocated by the caller.
///////////////////////////////////////////////////////////////
const static mic_gvf_soa      *mic_u      = NULL;
const static mic_clvf         *mic_fclinv = NULL;
const static mic_ras_gvf_soa  *ras_u      = NULL;
const static mic_ras_clvf     *ras_fclinv = NULL;

static bool is_mic_ras_fast_call = true;

///////////////////////////////////////////////////////////////
// mult counter 
///////////////////////////////////////////////////////////////
static int mic_ras_mult_counter = 0;
extern "C"
int get_mult_count() { return(mic_ras_mult_counter); }


extern "C" void
#ifdef _CLOVER
new_mic_ras_solver(const mic_nodeinfo    *ni, // node-infor passed from the offload interface
                   const mic_gvf_soa     *u,  // gauge field pointers passed from the offload interface
                   const mic_ras_gvf_soa *ru, // the memory must be allocated by the caller.
                   const mic_clvf        *fclinv,
                   const mic_ras_clvf    *rfclinv)
#else
new_mic_ras_solver(const mic_nodeinfo    *ni,
                   const mic_gvf_soa     *u,
                   const mic_ras_gvf_soa *ru)
#endif
////////////////////////////////////////////////////
// initialize MIC native RAS solver
////////////////////////////////////////////////////
{

  ////////////////////////////////
  // copy node infor
  ////////////////////////////////
  mic_nodeid_x = ni->mic_nodeid_x;
  mic_nodeid_y = ni->mic_nodeid_y;
  mic_nodeid_z = ni->mic_nodeid_z;
  mic_nodeid_x_up = ni->mic_nodeid_x_up;
  mic_nodeid_y_up = ni->mic_nodeid_y_up;
  mic_nodeid_z_up = ni->mic_nodeid_z_up;
  mic_nodeid_x_dn = ni->mic_nodeid_x_dn;
  mic_nodeid_y_dn = ni->mic_nodeid_y_dn;
  mic_nodeid_z_dn = ni->mic_nodeid_z_dn;
  mic_ipdeo  = ni->mic_ipdeo;
  mic_nodeid = ni->mic_nodeid;
  mic_ipeo   = ni->mic_ipeo;

  mic_ras_mult_counter = 0;


  if ( is_mic_ras_fast_call ) {
    if ( 0 == _nodeid_ ) {
      printf("%%==============================================================================\n");
#if   __MIC__ == 1
      printf("%% MIC Kernel is used\n");
#elif __AVX__ == 1
      printf("%% AVX Kernel is used\n");
#else
      printf("%%  No Kernel is used\n");
#endif
#ifdef _OPENMP
      printf("%% Number of OpenMP Threads : %d\n",omp_get_max_threads());
#else
      printf("%%            Single Thread : %d\n",1);
#endif
      printf("%%                 (NDIMX,NDIMY,NDIMZ) : (%3d,%3d,%3d)\n",_NDIMX,_NDIMY,_NDIMZ);

      printf("%%  Restricted Additive Shwartz method : overlap region depth\n");
      printf("%%           (NDEPTHX,NDEPTHY,NDEPTHZ) : (%3d,%3d,%3d)\n",NDEPTHX,NDEPTHY,NDEPTHZ);
      printf("%%                       RAS iteration : %3d\n",_NRAS);
      printf("%%                    Domain iteration : %3d\n",_NMR);
      printf("%%             SIZEOF(mic_wqf_eo) BYTE : %d\n",sizeof(mic_wqf_eo));
      printf("%%==============================================================================\n");
    fflush(stdout);
    }
    is_mic_ras_fast_call = false;
  }


  ////////////////////////////////
  // copy gauge field pointers
  ////////////////////////////////
  mic_u = u;
  ras_u = ru;
#ifdef _CLOVER
  mic_fclinv = fclinv;
  ras_fclinv = rfclinv;
#endif

  ////////////////////////////////////////////////////////////////////////////////
  // allocate and setup reverse-offloading communication interfaces
  //
  // {s,r}buff_{x,y,z}_{up,dn} : send/recv buffers for the nearest
  //                             neighbour surface boundary copy
  //
  // {s,r}{x,y,z}buff_{up,dn} : send/recv buffers for the overlapped region (halo)
  //                            boundary copy
  //
  ////////////////////////////////////////////////////////////////////////////////
#ifndef _singlePU
  void *p;
// get buffers for reverse offload of MPI calls
  scif_mic->allocbuf(2*sizeof(float), barb_send, p);
  arb_send = (float *)p;
  scif_mic->allocbuf(2*sizeof(float), barb_recv, p);
  arb_recv = (float *)p;
#if _NDIMZ != 1
  scif_mic->allocbuf(NRAS_GHOST_Z*sizeof(__mic_half_spinor_t), bs_z_dn, p);
  sbuff_z_dn = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(NRAS_GHOST_Z*sizeof(__mic_half_spinor_t), br_z_dn, p);
  rbuff_z_dn = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(NRAS_GHOST_Z*sizeof(__mic_half_spinor_t), bs_z_up, p);
  sbuff_z_up = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(NRAS_GHOST_Z*sizeof(__mic_half_spinor_t), br_z_up, p);
  rbuff_z_up = (__mic_half_spinor_t *)p;

  scif_mic->allocbuf(bsize_z, bszbuff_up, p);
  szbuff_up = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(bsize_z, brzbuff_up, p);
  rzbuff_up = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(bsize_z, bszbuff_dn, p);
  szbuff_dn = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(bsize_z, brzbuff_dn, p);
  rzbuff_dn = (__mic_half_spinor_t *)p;
#endif
#if _NDIMY != 1
  scif_mic->allocbuf(NRAS_GHOST_Y*sizeof(__mic_half_spinor_t), bs_y_dn, p);
  sbuff_y_dn = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(NRAS_GHOST_Y*sizeof(__mic_half_spinor_t), br_y_dn, p);
  rbuff_y_dn = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(NRAS_GHOST_Y*sizeof(__mic_half_spinor_t), bs_y_up, p);
  sbuff_y_up = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(NRAS_GHOST_Y*sizeof(__mic_half_spinor_t), br_y_up, p);
  rbuff_y_up = (__mic_half_spinor_t *)p;

  scif_mic->allocbuf(bsize_y, bsybuff_up, p);
  sybuff_up = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(bsize_y, brybuff_up, p);
  rybuff_up = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(bsize_y, bsybuff_dn, p);
  sybuff_dn = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(bsize_y, brybuff_dn, p);
  rybuff_dn = (__mic_half_spinor_t *)p;
#endif
#if _NDIMX != 1
  scif_mic->allocbuf(NRAS_GHOST_X*sizeof(__mic_half_spinor_t), bs_x_dn, p);
  sbuff_x_dn = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(NRAS_GHOST_X*sizeof(__mic_half_spinor_t), br_x_dn, p);
  rbuff_x_dn = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(NRAS_GHOST_X*sizeof(__mic_half_spinor_t), bs_x_up, p);
  sbuff_x_up = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(NRAS_GHOST_X*sizeof(__mic_half_spinor_t), br_x_up, p);
  rbuff_x_up = (__mic_half_spinor_t *)p;

  scif_mic->allocbuf(bsize_x, bsxbuff_up, p);
  sxbuff_up = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(bsize_x, brxbuff_up, p);
  rxbuff_up = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(bsize_x, bsxbuff_dn, p);
  sxbuff_dn = (__mic_half_spinor_t *)p;
  scif_mic->allocbuf(bsize_x, brxbuff_dn, p);
  rxbuff_dn = (__mic_half_spinor_t *)p;
#endif
#endif

#ifdef _TIMING_
  new_timer(&solver_timer);
#endif

}

extern "C" void
delete_mic_ras_solver()
////////////////////////////////////////////////////
// finalize MIC native RAS solver
////////////////////////////////////////////////////
{

  mic_ras_mult_counter = 0;

  mic_u = NULL;
  ras_u = NULL;
#ifdef _CLOVER
  mic_fclinv = NULL;
  ras_fclinv = NULL;
#endif

  #ifndef _singlePU
  scif_mic->clearbuf();
  #endif

#ifdef _TIMING_
  delete_timer(&solver_timer);
#endif
}

#endif
