#ifndef _MIC_WQF_ALGEBRA_H
#define _MIC_WQF_ALGEBRA_H

#include <mkl.h>


//template <typename T, int N>
//char (*count_helper(T (&array)[N]))[N];

//#define count(array)  sizeof(*count_helper(array))

constexpr size_t count(mic_wqf_eo *y) { return NTHTZYX; }
constexpr size_t count(mic_wqf    *y) { return NTHTZYX*2; }
constexpr size_t count(const mic_wqf_eo *y) { return NTHTZYX; }
constexpr size_t count(const mic_wqf    *y) { return NTHTZYX*2; }
constexpr size_t count(mic_ras_wqf_eo *y) { return NTHTZYX_RAS; }
constexpr size_t count(mic_ras_wqf    *y) { return NTHTZYX_RAS*2; }
constexpr size_t count(const mic_ras_wqf_eo *y) { return NTHTZYX_RAS; }
constexpr size_t count(const mic_ras_wqf    *y) { return NTHTZYX_RAS*2; }
constexpr size_t count(mic_sap_wqf_eo_deo *y) { return NTHTBZYX; }
constexpr size_t count(mic_sap_wqf_deo    *y) { return NTHTBZYX*2; }
constexpr size_t count(mic_sap_wqf        *y) { return NTHTBZYX*4; }
constexpr size_t count(const mic_sap_wqf_eo_deo *y) { return NTHTBZYX; }
constexpr size_t count(const mic_sap_wqf_deo    *y) { return NTHTBZYX*2; }
constexpr size_t count(const mic_sap_wqf        *y) { return NTHTBZYX*4; }

#define prod __prod_v1

#ifdef _TIMING_

#define _START_PERF_MON_ABS2(_a_) {\
  startTimer(solver_timer.abs2_timer);\
}

#define   _END_PERF_MON_ABS2(_a_) {\
  stopTimer(solver_timer.abs2_timer);\
  solver_timer.abs2_flop_count += QABS2*VLEN*count(_a_);\
  solver_timer.abs2_ld_dsize   += (double)sizeof((*_a_));\
}

#define _START_PERF_MON_PROD(_a_) {\
  startTimer(solver_timer.prod_timer);\
}

#define   _END_PERF_MON_PROD(_a_) {\
  stopTimer(solver_timer.prod_timer);\
  solver_timer.prod_flop_count += QPROD*VLEN*count(_a_);\
  solver_timer.prod_ld_dsize   += (double)(2*sizeof((*_a_)));\
}

#define _START_PERF_MON_REDUCT {\
  startTimer(solver_timer.reduct_timer);\
}
#define   _END_PERF_MON_REDUCT {\
  stopTimer(solver_timer.reduct_timer);\
}

#else

#define _START_PERF_MON_ABS2(_a_)
#define   _END_PERF_MON_ABS2(_a_)
#define _START_PERF_MON_PROD(_a_)
#define   _END_PERF_MON_PROD(_a_)

#define _START_PERF_MON_REDUCT
#define   _END_PERF_MON_REDUCT

#endif

#ifndef _singlePU
static inline void
mic_sumcast(float *f){
  _ALIG_PRE_ float f1, f2 _ALIG_PST_;
  f1 = *f;

  _START_PERF_MON_REDUCT

  mic_comm_allreduce(&f1,&f2,1);

  _END_PERF_MON_REDUCT

  *f = f2;
}

static inline void
mic_sumcast(float *cr, float *ci) {
  _ALIG_PRE_ float cc[2], ww[2] _ALIG_PST_;
  cc[0] = *cr;
  cc[1] = *ci;

  _START_PERF_MON_REDUCT

  mic_comm_allreduce(cc,ww,2);

  _END_PERF_MON_REDUCT

  *cr = ww[0];
  *ci = ww[1];
}
#endif

inline void
assign_attribute(mic_wqf_eo *a, const mic_wqf_eo *restrict b) 
{
  (*a).ieo = (*b).ieo;
}

inline void
assign_attribute(mic_wqf *a, const mic_wqf *restrict b) 
{
  (*a).eo[0].ieo = (*b).eo[0].ieo;
  (*a).eo[1].ieo = (*b).eo[1].ieo;
}

inline void
assign_attribute(mic_ras_wqf_eo *a, const mic_ras_wqf_eo *restrict b) 
{
  (*a).ieo = (*b).ieo;
}

inline void
assign_attribute(mic_ras_wqf *a, const mic_ras_wqf *restrict b) 
{
  (*a).eo[0].ieo = (*b).eo[0].ieo;
  (*a).eo[1].ieo = (*b).eo[1].ieo;
}

inline void
assign_attribute(mic_sap_wqf_eo_deo *a, const mic_sap_wqf_eo_deo *restrict b) 
{
  (*a).ideo = (*b).ideo;
  (*a).ieo  = (*b).ieo;
}

inline void
assign_attribute(mic_sap_wqf_deo *a, const mic_sap_wqf_deo *restrict b) 
{
  (*a).ideo = (*b).ideo;
  assign_attribute( &(*a).eo[0],  &(*b).eo[0]);
  assign_attribute( &(*a).eo[1],  &(*b).eo[1]);
}

inline void
assign_attribute(mic_sap_wqf *a, const mic_sap_wqf *restrict b) 
{
  assign_attribute( &(*a).deo[0],  &(*b).deo[0]);
  assign_attribute( &(*a).deo[1],  &(*b).deo[1]);
}

template <typename T>
inline void 
clear(T *a) 
{
//
// a = 0
//
#pragma omp parallel for
  for (int isite=0; isite < count(a); ++isite) {
    (*a).s[isite].y[0] = _my_set1_ps(0.0f);
    (*a).s[isite].y[1] = _my_set1_ps(0.0f);
    (*a).s[isite].y[2] = _my_set1_ps(0.0f);
    (*a).s[isite].y[3] = _my_set1_ps(0.0f);
    (*a).s[isite].y[4] = _my_set1_ps(0.0f);
    (*a).s[isite].y[5] = _my_set1_ps(0.0f);
  }
}

inline void 
clear(mic_sap_wqf_deo *a) 
{
  clear(&((*a).eo[0]));
  clear(&((*a).eo[1]));
}

inline void 
clear(mic_sap_wqf *a) 
{
  clear(&((*a).deo[0]));
  clear(&((*a).deo[1]));
}

template <typename T>
inline void 
assign(T *restrict a, const T *restrict b) 
{
//
// a = b
//
  (*a).ieo  = (*b).ieo;

#pragma omp parallel for
  for (int isite=0; isite < count(a); ++isite) {
    _my_ngostore_ps(&(*a).s[isite].y[0], (*b).s[isite].y[0]);
    _my_ngostore_ps(&(*a).s[isite].y[1], (*b).s[isite].y[1]);
    _my_ngostore_ps(&(*a).s[isite].y[2], (*b).s[isite].y[2]);
    _my_ngostore_ps(&(*a).s[isite].y[3], (*b).s[isite].y[3]);
    _my_ngostore_ps(&(*a).s[isite].y[4], (*b).s[isite].y[4]);
    _my_ngostore_ps(&(*a).s[isite].y[5], (*b).s[isite].y[5]);
  }

}

inline void 
assign(mic_sap_wqf_deo *restrict a, const mic_sap_wqf_deo *restrict b) 
{
  assign(&((*a).eo[0]),&((*b).eo[0]));
  assign(&((*a).eo[1]),&((*b).eo[1]));
}

inline void 
assign(mic_sap_wqf *restrict a, const mic_sap_wqf *restrict b) 
{
  assign(&((*a).deo[0]),&((*b).deo[0]));
  assign(&((*a).deo[1]),&((*b).deo[1]));
}


template <typename T>
inline float
__abs2_wo_comm(const T *a) 
{
//
// aba2 = |a|^2
//

  _START_PERF_MON_ABS2(a)

  _ALIG_PRE_ float aa _ALIG_PST_;

  aa = 0.0f;

#pragma omp parallel for reduction(+:aa)
  for (int isite=0; isite < count(a); ++isite) {

    register __mic_half_spinor_t aa1,aa2,aa3,aa4,aa5,aa6;

    aa1 = (*a).s[isite].y[0];
    aa2 = (*a).s[isite].y[1];
    aa3 = (*a).s[isite].y[2];
    aa4 = (*a).s[isite].y[3];
    aa5 = (*a).s[isite].y[4];
    aa6 = (*a).s[isite].y[5];

    aa1 =   _my_mul_ps(aa1,aa1);
    aa2 =   _my_mul_ps(aa2,aa2);
    aa1 = _my_fmadd_ps(aa3,aa3,aa1);
    aa2 = _my_fmadd_ps(aa4,aa4,aa2);
    aa1 = _my_fmadd_ps(aa5,aa5,aa1);
    aa2 = _my_fmadd_ps(aa6,aa6,aa2);

    aa1 = _my_add_ps(aa1,aa2);

    aa += _my_reduce_add_ps(aa1);

  }

  _END_PERF_MON_ABS2(a)

  return aa;

}

inline float
abs2(const mic_wqf_eo *a) 
{
  _ALIG_PRE_ float aa _ALIG_PST_;
  aa = __abs2_wo_comm(a);
#ifndef _singlePU
  mic_sumcast(&aa); 
#endif
  return aa;
}

inline float
abs2(const mic_ras_wqf_eo *a) 
{
  _ALIG_PRE_ float aa _ALIG_PST_;
  aa = __abs2_wo_comm(a);
#ifndef _singlePU
  mic_sumcast(&aa); 
#endif
  return aa;
}

inline float
abs2(const mic_wqf *a) 
{
  _ALIG_PRE_ float aa _ALIG_PST_;
  aa =  __abs2_wo_comm(&(a->eo[0]))
      + __abs2_wo_comm(&(a->eo[1]));
#ifndef _singlePU
  mic_sumcast(&aa); 
#endif
  return aa;
}

inline float
abs2(const mic_ras_wqf *a) 
{
  _ALIG_PRE_ float aa _ALIG_PST_;
  aa =  __abs2_wo_comm(&(a->eo[0]))
      + __abs2_wo_comm(&(a->eo[1]));
#ifndef _singlePU
  mic_sumcast(&aa); 
#endif
  return aa;
}

inline float
abs2(const mic_sap_wqf_eo_deo *a) 
{
  _ALIG_PRE_ float aa _ALIG_PST_;
  aa = __abs2_wo_comm(a);
#ifndef _singlePU
  mic_sumcast(&aa); 
#endif
  return aa;
}

inline float
abs2(const mic_sap_wqf_deo *a) 
{
  _ALIG_PRE_ float aa _ALIG_PST_;
  aa =  __abs2_wo_comm(&(a->eo[0]))
      + __abs2_wo_comm(&(a->eo[1]));
#ifndef _singlePU
  mic_sumcast(&aa); 
#endif
  return aa;
}

inline float
abs2(const mic_sap_wqf *a) 
{
  _ALIG_PRE_ float aa _ALIG_PST_;
  aa =  __abs2_wo_comm(&(a->deo[0].eo[0]))
      + __abs2_wo_comm(&(a->deo[0].eo[1]))
      + __abs2_wo_comm(&(a->deo[1].eo[0]))
      + __abs2_wo_comm(&(a->deo[1].eo[1]));
#ifndef _singlePU
  mic_sumcast(&aa); 
#endif
  return aa;
}

template <typename T>
inline void
__prod_wo_comm(float *p, const T *restrict a, const T *restrict b) 
{
//
// pp = <a|b>
//

  _START_PERF_MON_PROD(a)

  _ALIG_PRE_ float pr,pi _ALIG_PST_;

  pr = 0.0f;
  pi = 0.0f;

#pragma omp parallel for reduction(+:pr,pi)
  for (int isite=0; isite < count(a); ++isite) {
    register __mic_half_spinor_t pr1,pr2,pi1,pi2;
    register __mic_half_spinor_t aa1,aa2,aa3,aa4,aa5,aa6;
    register __mic_half_spinor_t bb1,bb2,bb3,bb4,bb5,bb6;
    register __mic_half_spinor_t neg = _my_set1_ps(-0.0f);

    aa1 = (*a).s[isite].y[0];
    aa2 = (*a).s[isite].y[1];
    aa3 = (*a).s[isite].y[2];
    aa4 = (*a).s[isite].y[3];
    aa5 = (*a).s[isite].y[4];
    aa6 = (*a).s[isite].y[5];

    bb1 = (*b).s[isite].y[0];
    bb2 = (*b).s[isite].y[1];
    bb3 = (*b).s[isite].y[2];
    bb4 = (*b).s[isite].y[3];
    bb5 = (*b).s[isite].y[4];
    bb6 = (*b).s[isite].y[5];

    pr1 =   _my_mul_ps(aa1,bb1);
    pr2 =   _my_mul_ps(aa2,bb2);
    pr1 = _my_fmadd_ps(aa3,bb3,pr1);
    pr2 = _my_fmadd_ps(aa4,bb4,pr2);
    pr1 = _my_fmadd_ps(aa5,bb5,pr1);
    pr2 = _my_fmadd_ps(aa6,bb6,pr2);

    pr1 = _my_add_ps(pr1,pr2);

    bb1 = _my_permute_BADC(bb1);
    bb2 = _my_permute_BADC(bb2);
    bb3 = _my_permute_BADC(bb3);
    bb4 = _my_permute_BADC(bb4);
    bb5 = _my_permute_BADC(bb5);
    bb6 = _my_permute_BADC(bb6);

    pi1 =   _my_mul_ps(aa1,bb1);
    pi2 =   _my_mul_ps(aa2,bb2);
    pi1 = _my_fmadd_ps(aa3,bb3,pi1);
    pi2 = _my_fmadd_ps(aa4,bb4,pi2);
    pi1 = _my_fmadd_ps(aa5,bb5,pi1);
    pi2 = _my_fmadd_ps(aa6,bb6,pi2);

    pi1 = _my_add_ps(pi1,pi2);
    pi1 = _my_xor_ps(pi1,MASK_FF00);

    pr += _my_reduce_add_ps(pr1);
    pi += _my_reduce_add_ps(pi1);

  }

  _END_PERF_MON_PROD(a)

  p[0] = pr;
  p[1] = pi;
}

inline std::complex<float>
prod(const mic_wqf_eo *restrict a, const mic_wqf_eo *restrict b) 
{

  assert(a->ieo == b->ieo);

  _ALIG_PRE_ float p[2] _ALIG_PST_;
  __prod_wo_comm(p,a,b);
#ifndef _singlePU
  mic_sumcast(&p[0],&p[1]); 
#endif
  return std::complex<float>(p[0],p[1]);
}

inline std::complex<float>
prod(const mic_wqf *restrict a, const mic_wqf *restrict b) {
  _ALIG_PRE_ float p0[2] _ALIG_PST_;
  _ALIG_PRE_ float p1[2] _ALIG_PST_;
  __prod_wo_comm(p0,&(a->eo[0]),&(b->eo[0]));
  __prod_wo_comm(p1,&(a->eo[1]),&(b->eo[1]));
  p0[0] += p1[0];
  p0[1] += p1[1];
#ifndef _singlePU
  mic_sumcast(&p0[0],&p0[1]); 
#endif
  return std::complex<float>(p0[0],p0[1]);
}

inline std::complex<float>
prod(const mic_sap_wqf_eo_deo *restrict a, const mic_sap_wqf_eo_deo *restrict b) {
  _ALIG_PRE_ float p[2] _ALIG_PST_;
  assert(a->ieo == b->ieo);
  assert(a->ideo == b->ideo);
  __prod_wo_comm(p,a,b);
#ifndef _singlePU
  mic_sumcast(&p[0],&p[1]); 
#endif
  return std::complex<float>(p[0],p[1]);
}

inline std::complex<float>
prod(const mic_sap_wqf_deo *restrict a, const mic_sap_wqf_deo *restrict b) {
  _ALIG_PRE_ float p0[2] _ALIG_PST_;
  _ALIG_PRE_ float p1[2] _ALIG_PST_;
  assert(a->ideo == b->ideo);
  __prod_wo_comm(p0,&(a->eo[0]),&(b->eo[0]));
  __prod_wo_comm(p1,&(a->eo[1]),&(b->eo[1]));
  p0[0] += p1[0];
  p0[1] += p1[1];
#ifndef _singlePU
  mic_sumcast(&p0[0],&p0[1]); 
#endif
  return std::complex<float>(p0[0],p0[1]);
}

inline std::complex<float>
prod(const mic_sap_wqf *restrict a, const mic_sap_wqf *restrict b) {
  _ALIG_PRE_ float p00[2] _ALIG_PST_;
  _ALIG_PRE_ float p01[2] _ALIG_PST_;
  _ALIG_PRE_ float p10[2] _ALIG_PST_;
  _ALIG_PRE_ float p11[2] _ALIG_PST_;
  __prod_wo_comm(p00,&(a->deo[0].eo[0]),&(b->deo[0].eo[0]));
  __prod_wo_comm(p01,&(a->deo[0].eo[1]),&(b->deo[0].eo[1]));
  __prod_wo_comm(p10,&(a->deo[1].eo[0]),&(b->deo[1].eo[0]));
  __prod_wo_comm(p11,&(a->deo[1].eo[1]),&(b->deo[1].eo[1]));
  p00[0] = p00[0] + p01[0] + p10[0] + p11[0];
  p00[1] = p00[1] + p01[1] + p10[1] + p11[1];
#ifndef _singlePU
  mic_sumcast(&p00[0],&p00[1]); 
#endif
  return std::complex<float>(p00[0],p00[1]);
}

template <typename T>
inline void
assign_add( T *restrict a, const T *restrict b, const T *restrict c){
//
// a = b + c
//
  (*a).ieo  = (*b).ieo;

#pragma omp parallel for
  for (int isite=0; isite < count(a); ++isite) {

    (*a).s[isite].y[0] = _my_add_ps((*b).s[isite].y[0], (*c).s[isite].y[0]);
    (*a).s[isite].y[1] = _my_add_ps((*b).s[isite].y[1], (*c).s[isite].y[1]);
    (*a).s[isite].y[2] = _my_add_ps((*b).s[isite].y[2], (*c).s[isite].y[2]);
    (*a).s[isite].y[3] = _my_add_ps((*b).s[isite].y[3], (*c).s[isite].y[3]);
    (*a).s[isite].y[4] = _my_add_ps((*b).s[isite].y[4], (*c).s[isite].y[4]);
    (*a).s[isite].y[5] = _my_add_ps((*b).s[isite].y[5], (*c).s[isite].y[5]);

  }
}

inline void
assign_add( mic_sap_wqf_deo *restrict a, const mic_sap_wqf_deo *restrict b, const mic_sap_wqf_deo *restrict c){
  assert(b->ideo == c->ideo);
  assign_add(&((*a).eo[0]),&((*b).eo[0]),&((*c).eo[0]));
  assign_add(&((*a).eo[1]),&((*b).eo[1]),&((*c).eo[1]));
}

inline void
assign_add( mic_sap_wqf *restrict a, const mic_sap_wqf *restrict b, const mic_sap_wqf *restrict c){
  assign_add(&((*a).deo[0]),&((*b).deo[0]),&((*c).deo[0]));
  assign_add(&((*a).deo[1]),&((*b).deo[1]),&((*c).deo[1]));
}

template <typename T>
inline void
assign_sub( T *restrict a, const T *restrict b, const T *restrict c){
//
// a = b - c
//
  (*a).ieo  = (*b).ieo;

#pragma omp parallel for
  for (int isite=0; isite < count(a); ++isite) {

    (*a).s[isite].y[0] = _my_sub_ps((*b).s[isite].y[0], (*c).s[isite].y[0]);
    (*a).s[isite].y[1] = _my_sub_ps((*b).s[isite].y[1], (*c).s[isite].y[1]);
    (*a).s[isite].y[2] = _my_sub_ps((*b).s[isite].y[2], (*c).s[isite].y[2]);
    (*a).s[isite].y[3] = _my_sub_ps((*b).s[isite].y[3], (*c).s[isite].y[3]);
    (*a).s[isite].y[4] = _my_sub_ps((*b).s[isite].y[4], (*c).s[isite].y[4]);
    (*a).s[isite].y[5] = _my_sub_ps((*b).s[isite].y[5], (*c).s[isite].y[5]);

  }
}

inline void
assign_sub( mic_sap_wqf_deo *restrict a, const mic_sap_wqf_deo *restrict b, const mic_sap_wqf_deo *restrict c){
  assert(b->ideo == c->ideo);
  assign_sub(&((*a).eo[0]),&((*b).eo[0]),&((*c).eo[0]));
  assign_sub(&((*a).eo[1]),&((*b).eo[1]),&((*c).eo[1]));
}

inline void
assign_sub( mic_sap_wqf *restrict a, const mic_sap_wqf *restrict b, const mic_sap_wqf *restrict c){
  assign_sub(&((*a).deo[0]),&((*b).deo[0]),&((*c).deo[0]));
  assign_sub(&((*a).deo[1]),&((*b).deo[1]),&((*c).deo[1]));
}

template <typename T>
inline void
accum_add( T *restrict w, const T *restrict y){
//
// w = w + y
//
#pragma omp parallel for
  for (int isite=0; isite < count(w); ++isite) {

    (*w).s[isite].y[0] = _my_add_ps((*w).s[isite].y[0], (*y).s[isite].y[0]);
    (*w).s[isite].y[1] = _my_add_ps((*w).s[isite].y[1], (*y).s[isite].y[1]);
    (*w).s[isite].y[2] = _my_add_ps((*w).s[isite].y[2], (*y).s[isite].y[2]);
    (*w).s[isite].y[3] = _my_add_ps((*w).s[isite].y[3], (*y).s[isite].y[3]);
    (*w).s[isite].y[4] = _my_add_ps((*w).s[isite].y[4], (*y).s[isite].y[4]);
    (*w).s[isite].y[5] = _my_add_ps((*w).s[isite].y[5], (*y).s[isite].y[5]);

  }
}

inline void
accum_add( mic_sap_wqf_deo *restrict w, const mic_sap_wqf_deo *restrict y){
  assert(w->ideo == y->ideo);
  accum_add(&((*w).eo[0]),&((*y).eo[0]));
  accum_add(&((*w).eo[1]),&((*y).eo[1]));
}

inline void
accum_add( mic_sap_wqf *restrict w, const mic_sap_wqf *restrict y){
  accum_add(&((*w).deo[0]),&((*y).deo[0]));
  accum_add(&((*w).deo[1]),&((*y).deo[1]));
}

template <typename T>
inline void
accum_sub( T *restrict w, const T *restrict y)
{
//
// w = w - y
//
#pragma omp parallel for
  for (int isite=0; isite < count(w); ++isite) {

    (*w).s[isite].y[0] = _my_sub_ps((*w).s[isite].y[0], (*y).s[isite].y[0]);
    (*w).s[isite].y[1] = _my_sub_ps((*w).s[isite].y[1], (*y).s[isite].y[1]);
    (*w).s[isite].y[2] = _my_sub_ps((*w).s[isite].y[2], (*y).s[isite].y[2]);
    (*w).s[isite].y[3] = _my_sub_ps((*w).s[isite].y[3], (*y).s[isite].y[3]);
    (*w).s[isite].y[4] = _my_sub_ps((*w).s[isite].y[4], (*y).s[isite].y[4]);
    (*w).s[isite].y[5] = _my_sub_ps((*w).s[isite].y[5], (*y).s[isite].y[5]);

  }
}

inline void
accum_sub( mic_sap_wqf_deo *restrict w, const mic_sap_wqf_deo *restrict y)
{
  assert(w->ideo == y->ideo);
  accum_sub(&((*w).eo[0]),&((*y).eo[0]));
  accum_sub(&((*w).eo[1]),&((*y).eo[1]));
}

inline void
accum_sub( mic_sap_wqf *restrict w, const mic_sap_wqf *restrict y)
{
  accum_sub(&((*w).deo[0]),&((*y).deo[0]));
  accum_sub(&((*w).deo[1]),&((*y).deo[1]));
}

template <typename T>
inline void
accum_mult_add( T *restrict w, const float *coef, const T *restrict y)
{
//
// w = w * coef + y
//
  register __mic_half_spinor_t cf = _my_load_bcast_ss(coef);

#pragma omp parallel for
  for (int isite=0; isite < count(w); ++isite) {

    (*w).s[isite].y[0] = _my_fmadd_ps((*w).s[isite].y[0], cf, (*y).s[isite].y[0]);
    (*w).s[isite].y[1] = _my_fmadd_ps((*w).s[isite].y[1], cf, (*y).s[isite].y[1]);
    (*w).s[isite].y[2] = _my_fmadd_ps((*w).s[isite].y[2], cf, (*y).s[isite].y[2]);
    (*w).s[isite].y[3] = _my_fmadd_ps((*w).s[isite].y[3], cf, (*y).s[isite].y[3]);
    (*w).s[isite].y[4] = _my_fmadd_ps((*w).s[isite].y[4], cf, (*y).s[isite].y[4]);
    (*w).s[isite].y[5] = _my_fmadd_ps((*w).s[isite].y[5], cf, (*y).s[isite].y[5]);

  }
}

template <typename T>
inline void
accum_mult_add( mic_sap_wqf_deo *restrict w, T *coef, const mic_sap_wqf_deo *restrict y)
{
  assert(w->ideo == y->ideo);
  accum_mult_add(&((*w).eo[0]),coef,&((*y).eo[0]));
  accum_mult_add(&((*w).eo[1]),coef,&((*y).eo[1]));
}

template <typename T>
inline void
accum_mult_add( mic_sap_wqf *restrict w, T *coef, const mic_sap_wqf *restrict y)
{
  accum_mult_add(&((*w).deo[0]),coef,&((*y).deo[0]));
  accum_mult_add(&((*w).deo[1]),coef,&((*y).deo[1]));
}

template <typename T>
inline void
accum_mult_add( T *restrict w, const std::complex<float> *coef, const T *restrict y)
{
//
// w = w * coef + y
//
  _ALIG_PRE_ const float (&a)[2] _ALIG_PST_ = reinterpret_cast<const float (&)[2]>(*coef);
  register __mic_half_spinor_t cfr = _my_load_bcast_ss(&a[0]);
  register __mic_half_spinor_t cfi = _my_load_bcast_ss(&a[1]);
  register __mic_half_spinor_t neg = _my_set1_ps(-0.0f);
  cfi = _my_xor_ps(cfi,MASK_00FF);

#pragma omp parallel for
  for (int isite=0; isite < count(w); ++isite) {
    register __mic_half_spinor_t w1,w2,w3,w4,w5,w6;
    register __mic_half_spinor_t s1,s2,s3,s4,s5,s6;
    register __mic_half_spinor_t neg = _my_set1_ps(-0.0f);

    w1 = (*w).s[isite].y[0];
    w2 = (*w).s[isite].y[1];
    w3 = (*w).s[isite].y[2];
    w4 = (*w).s[isite].y[3];
    w5 = (*w).s[isite].y[4];
    w6 = (*w).s[isite].y[5];

    s1 = _my_permute_BADC(w1);
    s2 = _my_permute_BADC(w2);
    s3 = _my_permute_BADC(w3);
    s4 = _my_permute_BADC(w4);
    s5 = _my_permute_BADC(w5);
    s6 = _my_permute_BADC(w6);

    w1 = _my_fmadd_ps(w1, cfr, (*y).s[isite].y[0]);
    w2 = _my_fmadd_ps(w2, cfr, (*y).s[isite].y[1]);
    w3 = _my_fmadd_ps(w3, cfr, (*y).s[isite].y[2]);
    w4 = _my_fmadd_ps(w4, cfr, (*y).s[isite].y[3]);
    w5 = _my_fmadd_ps(w5, cfr, (*y).s[isite].y[4]);
    w6 = _my_fmadd_ps(w6, cfr, (*y).s[isite].y[5]);

    (*w).s[isite].y[0] = _my_fmadd_ps(s1, cfi, w1);
    (*w).s[isite].y[1] = _my_fmadd_ps(s2, cfi, w2);
    (*w).s[isite].y[2] = _my_fmadd_ps(s3, cfi, w3);
    (*w).s[isite].y[3] = _my_fmadd_ps(s4, cfi, w4);
    (*w).s[isite].y[4] = _my_fmadd_ps(s5, cfi, w5);
    (*w).s[isite].y[5] = _my_fmadd_ps(s6, cfi, w6);

  }
}

template <typename T>
inline void
accum_add_mult( T *restrict w, const float *coef, const T *restrict y)
{
//
// w = w + coef * y
//
  register __mic_half_spinor_t cf = _my_load_bcast_ss(coef);

#pragma omp parallel for
  for (int isite=0; isite < count(w); ++isite) {

    (*w).s[isite].y[0] = _my_fmadd_ps((*y).s[isite].y[0], cf, (*w).s[isite].y[0]);
    (*w).s[isite].y[1] = _my_fmadd_ps((*y).s[isite].y[1], cf, (*w).s[isite].y[1]);
    (*w).s[isite].y[2] = _my_fmadd_ps((*y).s[isite].y[2], cf, (*w).s[isite].y[2]);
    (*w).s[isite].y[3] = _my_fmadd_ps((*y).s[isite].y[3], cf, (*w).s[isite].y[3]);
    (*w).s[isite].y[4] = _my_fmadd_ps((*y).s[isite].y[4], cf, (*w).s[isite].y[4]);
    (*w).s[isite].y[5] = _my_fmadd_ps((*y).s[isite].y[5], cf, (*w).s[isite].y[5]);

  }
}

template <typename T>
inline void
accum_add_mult( mic_sap_wqf_deo *restrict w, T *coef, const mic_sap_wqf_deo *restrict y)
{
  assert(w->ideo == y->ideo);
  accum_add_mult(&((*w).eo[0]),coef,&((*y).eo[0]));
  accum_add_mult(&((*w).eo[1]),coef,&((*y).eo[1]));
}

template <typename T>
inline void
accum_add_mult( mic_sap_wqf *restrict w, T *coef, const mic_sap_wqf *restrict y)
{
  accum_add_mult(&((*w).deo[0]),coef,&((*y).deo[0]));
  accum_add_mult(&((*w).deo[1]),coef,&((*y).deo[1]));
}

template <typename T>
inline void
accum_sub_mult( T *restrict w, const float *coef, const T *restrict y)
{
//
// w = w - coef * y
//
  register __mic_half_spinor_t cf = _my_load_bcast_ss(coef);

#pragma omp parallel for
  for (int isite=0; isite < count(w); ++isite) {

    (*w).s[isite].y[0] = _my_fnmadd_ps((*y).s[isite].y[0], cf, (*w).s[isite].y[0]);
    (*w).s[isite].y[1] = _my_fnmadd_ps((*y).s[isite].y[1], cf, (*w).s[isite].y[1]);
    (*w).s[isite].y[2] = _my_fnmadd_ps((*y).s[isite].y[2], cf, (*w).s[isite].y[2]);
    (*w).s[isite].y[3] = _my_fnmadd_ps((*y).s[isite].y[3], cf, (*w).s[isite].y[3]);
    (*w).s[isite].y[4] = _my_fnmadd_ps((*y).s[isite].y[4], cf, (*w).s[isite].y[4]);
    (*w).s[isite].y[5] = _my_fnmadd_ps((*y).s[isite].y[5], cf, (*w).s[isite].y[5]);

  }
}

template <typename T>
inline void
accum_sub_mult( mic_sap_wqf_deo *restrict w, T *coef, const mic_sap_wqf_deo *restrict y)
{
  accum_sub_mult(&((*w).eo[0]),coef,&((*y).eo[0]));
  accum_sub_mult(&((*w).eo[1]),coef,&((*y).eo[1]));
}

template <typename T>
inline void
accum_sub_mult( mic_sap_wqf *restrict w, T *coef, const mic_sap_wqf *restrict y)
{
  accum_sub_mult(&((*w).deo[0]),coef,&((*y).deo[0]));
  accum_sub_mult(&((*w).deo[1]),coef,&((*y).deo[1]));
}

template <typename T>
inline void
accum_add_mult( T *restrict w, const std::complex<float> *coef, const T *restrict y)
{
//
// w = w + coef * y
//
  _ALIG_PRE_ const float (&a)[2] _ALIG_PST_ = reinterpret_cast<const float (&)[2]>(*coef);
  register __mic_half_spinor_t cfr = _my_load_bcast_ss(&a[0]);
  register __mic_half_spinor_t cfi = _my_load_bcast_ss(&a[1]);
  register __mic_half_spinor_t neg = _my_set1_ps(-0.0f);
  cfi = _my_xor_ps(cfi,MASK_00FF);

#pragma omp parallel for
  for (int isite=0; isite < count(w); ++isite) {
    register __mic_half_spinor_t w1,w2,w3,w4,w5,w6;
    register __mic_half_spinor_t s1,s2,s3,s4,s5,s6;
    register __mic_half_spinor_t neg = _my_set1_ps(-0.0f);

    w1 = (*y).s[isite].y[0];
    w2 = (*y).s[isite].y[1];
    w3 = (*y).s[isite].y[2];
    w4 = (*y).s[isite].y[3];
    w5 = (*y).s[isite].y[4];
    w6 = (*y).s[isite].y[5];

    s1 = _my_permute_BADC(w1);
    s2 = _my_permute_BADC(w2);
    s3 = _my_permute_BADC(w3);
    s4 = _my_permute_BADC(w4);
    s5 = _my_permute_BADC(w5);
    s6 = _my_permute_BADC(w6);

    w1 = _my_fmadd_ps(w1, cfr, (*w).s[isite].y[0]);
    w2 = _my_fmadd_ps(w2, cfr, (*w).s[isite].y[1]);
    w3 = _my_fmadd_ps(w3, cfr, (*w).s[isite].y[2]);
    w4 = _my_fmadd_ps(w4, cfr, (*w).s[isite].y[3]);
    w5 = _my_fmadd_ps(w5, cfr, (*w).s[isite].y[4]);
    w6 = _my_fmadd_ps(w6, cfr, (*w).s[isite].y[5]);

    (*w).s[isite].y[0] = _my_fmadd_ps(s1, cfi, w1);
    (*w).s[isite].y[1] = _my_fmadd_ps(s2, cfi, w2);
    (*w).s[isite].y[2] = _my_fmadd_ps(s3, cfi, w3);
    (*w).s[isite].y[3] = _my_fmadd_ps(s4, cfi, w4);
    (*w).s[isite].y[4] = _my_fmadd_ps(s5, cfi, w5);
    (*w).s[isite].y[5] = _my_fmadd_ps(s6, cfi, w6);

  }
}

template <typename T>
inline void
accum_sub_mult( T *restrict w, const std::complex<float> *coef, const T *restrict y)
{
//
// w = w - coef * y
//
  _ALIG_PRE_ const float (&a)[2] _ALIG_PST_ = reinterpret_cast<const float (&)[2]>(*coef);
  register __mic_half_spinor_t cfr = _my_load_bcast_ss(&a[0]);
  register __mic_half_spinor_t cfi = _my_load_bcast_ss(&a[1]);
  register __mic_half_spinor_t neg = _my_set1_ps(-0.0f);
  cfi = _my_xor_ps(cfi,MASK_00FF);

#pragma omp parallel for
  for (int isite=0; isite < count(w); ++isite) {
    register __mic_half_spinor_t w1,w2,w3,w4,w5,w6;
    register __mic_half_spinor_t s1,s2,s3,s4,s5,s6;
    register __mic_half_spinor_t neg = _my_set1_ps(-0.0f);

    w1 = (*y).s[isite].y[0];
    w2 = (*y).s[isite].y[1];
    w3 = (*y).s[isite].y[2];
    w4 = (*y).s[isite].y[3];
    w5 = (*y).s[isite].y[4];
    w6 = (*y).s[isite].y[5];

    s1 = _my_permute_BADC(w1);
    s2 = _my_permute_BADC(w2);
    s3 = _my_permute_BADC(w3);
    s4 = _my_permute_BADC(w4);
    s5 = _my_permute_BADC(w5);
    s6 = _my_permute_BADC(w6);

    w1 = _my_fnmadd_ps(w1, cfr, (*w).s[isite].y[0]);
    w2 = _my_fnmadd_ps(w2, cfr, (*w).s[isite].y[1]);
    w3 = _my_fnmadd_ps(w3, cfr, (*w).s[isite].y[2]);
    w4 = _my_fnmadd_ps(w4, cfr, (*w).s[isite].y[3]);
    w5 = _my_fnmadd_ps(w5, cfr, (*w).s[isite].y[4]);
    w6 = _my_fnmadd_ps(w6, cfr, (*w).s[isite].y[5]);

    (*w).s[isite].y[0] = _my_fnmadd_ps(s1, cfi, w1);
    (*w).s[isite].y[1] = _my_fnmadd_ps(s2, cfi, w2);
    (*w).s[isite].y[2] = _my_fnmadd_ps(s3, cfi, w3);
    (*w).s[isite].y[3] = _my_fnmadd_ps(s4, cfi, w4);
    (*w).s[isite].y[4] = _my_fnmadd_ps(s5, cfi, w5);
    (*w).s[isite].y[5] = _my_fnmadd_ps(s6, cfi, w6);

  }
}

#endif
