#include <config.h>

program main_quarks
!
! = Solve Quark propagator for Hadrons
!
  use comlib
  use file_tools_class
  use error_class
  use lattice_class
  use field_gauge_class
  use timer_class
  use wavefunc_class
  use quark_prop_class
  use hadron_class
  use main_hadron_class
#ifdef _STOUT
  use hmc_stout_class
#endif
  implicit none
  character(len=CHARLEN), parameter :: &
 &   prog_ver="$Id: main_hadron.F90,v 1.1 2006/10/04 01:52:25 ishikawa Exp $"

  type(main_hadron_obj) :: main

  type(quark_prop), allocatable :: qprop(:,:)
  type(snk_wavefunc_obj), allocatable, target :: snk_wf
#ifdef _STOUT
  type(stout_smearing_parameters) :: stout
#endif

  integer :: iout
  integer :: traj,mu
  real(DP) :: plq0
  character(len=CHARLEN) :: char
  character(len=CHARLEN) :: str
  integer :: isrc,iqk,isrc1,isrc2,isrcd,iflv
  integer :: ifm,ifb

  integer :: nfl,nfh,nf,iqk1,iqk2

  call new(main)

  if (nodeid == 0) then
    write(*,'(80("-"))')
    write(*,'(1X,A," for ",A)')TRIM(prog_ver),TRIM(_VERSION_)
    write(*,'(80("-"))')
  endif

  iout = search_free_file_unit()
  call read(main,iout)
  close(iout)
  call print(main)

  allocate(snk_wf)
  call new(snk_wf)
  snk_wf%param = main%wavefunc(1)%param
  call get(snk_wf)

  allocate(qprop(2,2))
  do isrc=1,2
  do iqk=1,2
    call new(qprop(iqk,isrc))
  enddo
  enddo

  do traj=main%ini_traj,main%end_traj,main%traj_skip

    !-----------------------
    ! read config
    !-----------------------
    call read(traj,main%config_path,main%config_name, &
 &                 main%config_prof,main%u)

#ifdef _STOUT
    if (main%switch_stout == SW_ON) then
      call new(stout,alpha=main%stout_alpha,nstout=main%stout_nstep,is_smear_only=.TRUE.)
      call print(stout)
      call smear(stout,main%u)
      call assign(main%u,stout%usmr(1))
      call delete(stout)
      call copy_boundary(main%u)
      call plaquette(main%u,plq0)
      if (nodeid==0) write(*,'(" PLQ(STOUT):",F24.16)')plq0
    endif
#endif


    !-----------------------
    ! create quark props
    !-----------------------
    do isrc = 1,2
    do iqk  = 1,2
      qprop(iqk,isrc)%parameter     => main%quark_parameter(iqk)

      select case(main%num_wave)
      case (3) ! independent strange quark wavefunction and up/down quark wavefunction
        if ( iqk == 1 ) then
          if ( isrc == 1 ) then
            qprop(iqk,isrc)%wavefunc_sous => main%wavefunc(1)
          else if ( isrc == 2 ) then
            qprop(iqk,isrc)%wavefunc_sous => main%wavefunc(2)
          endif
        else if ( iqk == 2 ) then
          if ( isrc == 1 ) then
            qprop(iqk,isrc)%wavefunc_sous => main%wavefunc(1)
          else if ( isrc == 2 ) then
            qprop(iqk,isrc)%wavefunc_sous => main%wavefunc(3) ! strange quark smeared wavefunction
          endif
        endif
      case (2)
        qprop(iqk,isrc)%wavefunc_sous => main%wavefunc(isrc)
      end select

      qprop(iqk,isrc)%smear_sous    =  isrc-1

      qprop(iqk,isrc)%wavefunc_sink => snk_wf
      qprop(iqk,isrc)%smear_sink    =  0

      if (nodeid==0) then
        write(*,'("")')
        write(*,'(80("="))')
        write(*,'("Begin ==== Quark #",I3," Source #",I3)')iqk,isrc
      endif
      call print(qprop(iqk,isrc)%parameter)
      if (nodeid==0) then
        write(*,'("")')
      endif

      !===============================
      ! generate source wavefunction
      !===============================
      call print(qprop(iqk,isrc)%wavefunc_sous)
      call   get(qprop(iqk,isrc)%wavefunc_sous)

      !===================================================
      ! compute smared source-local sink quark propagator
      !===================================================
      call set_source(qprop(iqk,isrc))
      call get_prop(qprop(iqk,isrc),main%u)
      deallocate(qprop(iqk,isrc)%wavefunc_sous%wavefunc)


      if (nodeid==0) then
        write(*,'("")')
      endif
      call print_statistics(qprop(iqk,isrc))
      if (nodeid==0) then
        write(*,'("End ==== Quark #",I3," Source #",I3)')iqk,isrc
        write(*,'("")')
      endif

      !===================
      ! Store qprop
      !===================
      write(str,'("qprop.",I2.2,".",I2.2,".",I6.6)')iqk,isrc,traj
      call store(qprop(iqk,isrc),main%config_prof%plq,"./",TRIM(str))

    enddo
    enddo

  enddo

  do isrc = 1,2
  do iqk  = 1,2
    call delete(qprop(iqk,isrc))
  enddo
  enddo
  deallocate(qprop)

  call delete(main)

  stop
end program
