#$Id: make.inc,v 1.4 2011/10/29 09:06:33 ishikawa Exp $
#########################################################################
### include file for Makefile
### COMA version
#########################################################################
DATE = '"$(shell env LC_ALL=C date +"%b %d %Y")"'
TIME = '"$(shell env LC_ALL=C date +"%H:%M:%S")"'
REVISION = -D_REVISION_='"$(shell cat $(TOPDIR)/.svnrevision)"'

#####################################################################
# Common section
#####################################################################
RM = /bin/rm -f
MV = /bin/mv
TOPDIR := $(dir $(lastword $(MAKEFILE_LIST)))
TOOLSDIR := $(TOPDIR)/Tools
MKDEPF90 := $(TOOLSDIR)/mkdepf90.perl

#####################################################################
# Compiler Secion
#####################################################################
MACHINE := INTEL

###################################
# Intel Fortran
###################################
ifeq ($(MACHINE),INTEL)

###################
# Archiver
###################
AR      = xiar
RANLIB  = ranlib

###################
# LAPACK95 PATH
###################
LAPACKINC = -I$(HOME)/include/lapack95_modules_intel
LAPACKLIB = -L$(HOME)/lib -llapack95_intel

###################
# LAPACK PATH 
# uses Inetl MKL
###################
LAPACKLIB += -mkl 

#################
# FFT LIBS
#################
FFTINC = -I$(HOME)/include
FFTLIB = $(HOME)/lib/libfftw3_intel.a
FFTLIB = $(HOME)/lib/libfftw3_omp.a

###################
# MPI compilers
###################
FC      = mpiifort
CC      = mpiicc 
CXX     = mpiicpc
CPP     = cpp

MIC_TBBLIB_PATH = /app/intel/composer_xe_2015.3.187/tbb/lib/mic


MICDIR = ../MICSolverClass

OPENMP  = -openmp -openmp-report2

#DEBUG   = -g -traceback -D_FLAT_ -debug all -check all
#DEBUG   = -g -traceback -D_FLAT_ -debug all -check all -D_DEBUG_
#DEBUG   = -g -traceback -D_FLAT_ -debug all -D_DEBUG_
DEBUG   = -g -traceback -D_FLAT_ -debug all


COMAINC = -I$(MICDIR)
COMALIB = -lstdc++ -L$(MICDIR) -lmicoffload -lscif \
          -offload-option,mic,ld,$(MICDIR)/libmicnative.a \
          -offload-option,mic,ld,-L$(MIC_TBBLIB_PATH) \
          -offload-option,mic,ld,-ltbbmalloc_proxy \
          -offload-option,mic,ld,-ltbbmalloc \
          -offload-option,mic,ld,-limf \
          -offload-option,mic,ld,-lsvml \
          -offload-option,mic,ld,-lintlc \
          -offload-option,mic,ld,-lirng \
          -offload-option,mic,ld,-lscif 

#          -lchkp -lchkpwrap -offload-option,mic,ld,-lchkp -offload-option,mic,ld,-lchkpwrap
#          -offload-option,mic,ld,-ltbbmalloc_proxy \
#          -offload-option,mic,ld,-ltbbmalloc \
#          -offload-option,mic,ld,-lscif

MICSOLVER = -D_MIC_SOLVER_

FFLAGS  = -Ofast -stand f03 $(MICSOLVER) \
          -qno-offload  -align array64byte \
          -finline -finline-limit=100000 -inline-factor=10000 -override-limits \
          $(DEBUG) $(OPENMP)

#          -shared-intel -align all \

LDFLAGS = $(FFLAGS)

endif


#####################################################################
# Summay section
#####################################################################

LIBS     = $(LAPACKLIB) $(FFTLIB) $(COMALIB) $(SYSLIB)
INCLUDES = $(LAPACKINC) $(FFTINC) $(COMAINC) $(SYSINC) \
           $(shell ls -1 -d ../*Class ../*Module | awk '{print "-I"$$1" "}')

OBJS    = $(SOURCES:.F90=.o)
LOGS    = $(SOURCES:.F90=.diag) $(SOURCES:.F90=.L) \
          $(SOURCES:.F90=.par)  $(SOURCES:.F90=.log) \
          $(SOURCES:.F90=.lst)

.PHONY: clean

.F90.o :

%.o : %.F90
	$(FC) $(FFLAGS) $(REVISION) $(MONITOR) $(INCLUDES) -c $< -o $@

%.s : %.F90
	$(FC) $(FFLAGS) $(REVISION) $(MONITOR) $(INCLUDES) -S $< -o $@

#########################################################################
